#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from math import asin
from typing import Dict, Iterable, List, Set, Tuple

import numpy as np
import pandas as pd
import scipy.stats as sps
import statsmodels.stats as stats
import statsmodels.stats.api as sms

import ambrosia.tools.pvalue_tools as pvalue_pkg
from ambrosia import types
from ambrosia.tools.configs import Alternatives

from . import EFFECT_COL_NAME, FIRST_TYPE_ERROR_COL_NAME, GROUP_SIZE_COL_NAME, STAT_ERRORS_COL_NAME

FIRST_TYPE_ERROR: float = 0.05
SECOND_TYPE_ERROR: float = 0.2
EFFECT_DEFAULT: float = 1.01
ROUND_DIGITS_TABLE: int = 3
ROUND_DIGITS_PERCENT: int = 1


def get_stats(values: Iterable[float], ddof: int = 1) -> Tuple[float, float]:
    """
    Calculate the mean and standard value for a list of values.
    """
    return np.mean(values), np.std(values, ddof=ddof)  # 1 for unbiased estimation


def get_table_stats(data: pd.DataFrame, column: types.ColumnNameType) -> Tuple[float, float]:
    """
    Calculate the mean and standard value a data frame column.
    """
    return get_stats(data[column].values)


def check_encode_alternative(alternative: str) -> str:
    """
    Check the correctness of the alternative and encode for use in statsmodels api.
    """
    alternatives: Set = {"two-sided", "greater", "less"}
    statsmodels_alternatives_encoding: Dict = {"two-sided": "two-sided", "greater": "larger", "less": "smaller"}
    if alternative not in alternatives:
        raise ValueError(f"Alternative must be one of '{alternatives}'.")
    else:
        return statsmodels_alternatives_encoding[alternative]


def unbiased_to_sufficient(std: float, size: int) -> float:
    """
    Transforms unbiased estimation of standard deviation to sufficient
    (ddof = 1) => (ddof = 0)
    """
    return std * np.sqrt((size - 1) / size)


def check_target_type(
    dataframe: pd.DataFrame,
    column: types.ColumnNameType,
):
    """
    Check type of target: binary / non-binary.
    """
    unique_vals_count = dataframe[column].nunique()
    if unique_vals_count > 2:
        return "non-binary"
    unique_vals = set(dataframe[column].unique())
    if unique_vals == {0, 1} or unique_vals in {0, 1}:
        return "binary"
    return "non-binary"


def stabilize_effect(
    eff: float, mean: float, std: float, target_type: str = "binary", stabilizing_method: str = "asin"
):
    """
    Evaluate stabilized effect for solve_power method.
    """
    if target_type == "non-binary":
        return (eff - 1) * mean / std
    elif target_type == "binary":
        if stabilizing_method == "asin":
            return 2 * (asin((mean * eff) ** 0.5) - asin(mean**0.5))
        elif stabilizing_method == "norm":
            return mean * (eff - 1) / ((mean * (1 - mean)) ** 0.5)
        else:
            raise Exception("Invalid stabilizing_method")
    else:
        raise Exception("Invalid target_type")


def destabilize_effect(
    eff: float, mean: float, std: float, target_type: str = "binary", stabilizing_method: str = "asin"
):
    """
    Evaluate destabilized effect from solve_power method (statsmodels).
    """
    if target_type == "non-binary":
        return eff * std / mean
    elif target_type == "binary":
        if stabilizing_method == "asin":
            return np.sin((eff + 2 * asin(mean**0.5)) / 2) ** 2 / mean - 1
        elif stabilizing_method == "norm":
            return eff * (mean * (1 - mean)) ** 0.5 / mean
        else:
            raise Exception("Invalid stabilizing_method")
    else:
        raise Exception("Invalid target_type")


def get_sample_size(
    mean: float,
    std: float,
    eff: float = EFFECT_DEFAULT,
    alpha: float = FIRST_TYPE_ERROR,
    beta: float = SECOND_TYPE_ERROR,
    target_type: str = "non-binary",
    groups_ratio: float = 1.0,
    alternative: str = "two-sided",
    stabilizing_method: str = "asin",
):
    """
    Calculate minimum sample size to catch effect with fixed errors.

    Parameters
    ----------
    mean : float
        Sample mean
    std : float
        Sample standard deviation
    eff : float, default: ``1.01``
        Effect for which we calculate sample size
    alpha : float, default: ``0.05``
        First type error
    beta : float, default: ``0.2``
        Second type error
    target_type : str, default: ``non-binary``
        Type of target metric: binary or non-binary
    groups_ratio : float, default: ``1.0``
        Ratio between two groups.
    alternative : str, default: ``"two-sided"``
        Alternative hypothesis, can be ``"two-sided"``, ``"greater"``
        or ``"less"``.
        ``"greater"`` - if effect is positive.
        ``"less"`` - if effect is negative.
    stabilizing_method : str, default: ``"asin"``
        Effect trasformation. Can be ``"asin"`` and ``"norm"``.
        For non-binary metrics: only ``"norm"`` is accceptable.
        For binary metrics: ``"norm"`` and ``"asin"``, but ``"asin"``
        is more robust and accurate.

    Returns
    -------
    sample_size : int
        Minimal sample size

    """
    alternative: str = check_encode_alternative(alternative)
    power_class = stats.power.TTestIndPower() if target_type == "non-binary" else stats.power.NormalIndPower()
    stabilized_effect = stabilize_effect(
        eff=eff, mean=mean, std=std, target_type=target_type, stabilizing_method=stabilizing_method
    )
    sample_size = power_class.solve_power(
        effect_size=stabilized_effect,
        nobs1=None,
        alpha=alpha,
        power=1 - beta,
        ratio=groups_ratio,
        alternative=alternative,
    )
    return int(np.ceil(sample_size))


def get_minimal_determinable_effect(
    mean: float,
    std: float,
    sample_size: int,
    alpha: float = FIRST_TYPE_ERROR,
    beta: float = SECOND_TYPE_ERROR,
    target_type: str = "non-binary",
    groups_ratio: float = 1.0,
    alternative: str = "two-sided",
    stabilizing_method: str = "asin",
):
    """
    Calculate power for given minimum detectable effect and group size.

    Parameters
    ----------
    mean : float
        Sample mean
    std : float
        Sample standard deviation
    sample_size : int
        Size of sample
    alpha : float, default: ``0.05``
        First type error
    beta : float, default: ``0.2``
        Second type error
    target_type : str, default: ``non-binary``
        Type of target metric: binary or non-binary
    groups_ratio : float, default: ``1.0``
        Ratio between two groups.
    alternative : str, default: ``"two-sided"``
        Alternative hypothesis, can be ``"two-sided"``, ``"greater"``
        or ``"less"``.
        ``"greater"`` - if effect is positive.
        ``"less"`` - if effect is negative.
    stabilizing_method : str, default: ``"asin"``
        Effect trasformation. Can be ``"asin"`` and ``"norm"``.
        For non-binary metrics: only ``"norm"`` is accceptable.
        For binary metrics: ``"norm"`` and ``"asin"``, but ``"asin"``
        is more robust and accurate.

    Returns
    -------
    mde : float
        Minimal effect which we can find

    """
    alternative: str = check_encode_alternative(alternative)
    power_class = stats.power.TTestIndPower() if target_type == "non-binary" else stats.power.NormalIndPower()
    stabilized_mde = power_class.solve_power(
        effect_size=None,
        nobs1=sample_size,
        alpha=alpha,
        power=1 - beta,
        ratio=groups_ratio,
        alternative=alternative,
    )
    mde = destabilize_effect(stabilized_mde, mean, std, target_type, stabilizing_method)
    return mde


def get_power(
    mean: float,
    std: float,
    sample_size: int,
    effect: float,
    alpha: float = FIRST_TYPE_ERROR,
    target_type: str = "non-binary",
    groups_ratio: float = 1.0,
    alternative: str = "two-sided",
    stabilizing_method: str = "asin",
) -> float:
    """
    Calculate minimum detectable effect which we can find.

    Parameters
    ----------
    mean : float
        Sample mean
    std : float
        Sample standard deviation
    sample_size : int
        Size of sample
    effect : float
        Second type error
    alpha : float, default: ``1.01``
        First type error
    target_type : str, default: ``non-binary``
        Type of target metric: binary or non-binary
    groups_ratio : float, default: ``1.0``
        Ratio between two groups.
    alternative : str, default: ``"two-sided"``
        Alternative hypothesis, can be ``"two-sided"``, ``"greater"``
        or ``"less"``.
        ``"greater"`` - if effect is positive.
        ``"less"`` - if effect is negative.
    stabilizing_method : str, default: ``"asin"``
        Effect trasformation. Can be ``"asin"`` and ``"norm"``.
        For non-binary metrics: only ``"norm"`` is accceptable.
        For binary metrics: ``"norm"`` and ``"asin"``, but ``"asin"``
        is more robust and accurate.

    Returns
    -------
    power : float
        Power effect with fixed size and effect
    """
    alternative: str = check_encode_alternative(alternative)
    power_class = stats.power.TTestIndPower() if target_type == "non-binary" else stats.power.NormalIndPower()
    stabilized_effect = stabilize_effect(
        eff=effect, mean=mean, std=std, target_type=target_type, stabilizing_method=stabilizing_method
    )
    power = power_class.solve_power(
        effect_size=stabilized_effect,
        nobs1=sample_size,
        alpha=alpha,
        power=None,
        ratio=groups_ratio,
        alternative=alternative,
    )
    return power


def get_table_sample_size(
    mean: float,
    std: float,
    effects: types.EffectType,
    first_errors: types.StatErrorType = (0.05,),
    second_errors: types.StatErrorType = (0.2,),
    target_type: str = "non-binary",
    groups_ratio: float = 1.0,
    alternative: str = "two-sided",
    stabilizing_method: str = "asin",
):
    """
    Create table of sample sizes for different effects and errors.

    Parameters
    ----------
    mean : float
        Sample mean
    std : float
        Sample standard deviation.
    effects : List
        List of effects which we want to catch.
        e.x.: [1.01, 1.02, 1.05]
    first_errors : List
        1st and 2nd type errors.
        e.x.: [0.01, 0.05, 0.1]
    second_errors : List
        1st and 2nd type errors.
        e.x.: [0.01, 0.05, 0.1]
    target_type : str, default: ``non-binary``
        Type of target metric: binary or non-binary
    groups_ratio : float, default: ``1.0``
        Ratio between two groups.
    alternative : str, default: ``"two-sided"``
        Alternative hypothesis, can be ``"two-sided"``, ``"greater"``
        or ``"less"``.
        ``"greater"`` - if effect is positive.
        ``"less"`` - if effect is negative.
    stabilizing_method : str, default: ``"asin"``
        Effect trasformation. Can be ``"asin"`` and ``"norm"``.
        For non-binary metrics: only ``"norm"`` is accceptable.
        For binary metrics: ``"norm"`` and ``"asin"``, but ``"asin"``
        is more robust and accurate.

    Returns
    -------
    df_results : pd.DataFrame
        Table with minimal sample sizes for each effect and error from input data.
    """
    multiindex = pd.MultiIndex.from_tuples([(eff,) for eff in effects], names=[EFFECT_COL_NAME])
    multicols = pd.MultiIndex.from_tuples(
        [(f"({err_one}; {err_two})",) for err_one in first_errors for err_two in second_errors],
        names=[STAT_ERRORS_COL_NAME],
    )
    df_results = pd.DataFrame(index=multiindex, columns=multicols)

    for eff in effects:
        for first_err in first_errors:
            for second_err in second_errors:
                err = f"({first_err}; {second_err})"
                df_results.loc[(eff,), (err,)] = get_sample_size(
                    mean=mean,
                    std=std,
                    eff=eff,
                    alpha=first_err,
                    beta=second_err,
                    target_type=target_type,
                    groups_ratio=groups_ratio,
                    alternative=alternative,
                    stabilizing_method=stabilizing_method,
                )
    df_results.index = pd.MultiIndex(
        levels=[[f"{np.round((x - 1) * 100, ROUND_DIGITS_PERCENT)}%" for x in effects]],
        codes=[np.arange(len(effects))],
        names=[EFFECT_COL_NAME],
    )
    return df_results


def design_groups_size(
    dataframe: pd.DataFrame,
    column: types.ColumnNameType,
    effects: Iterable[float],
    first_errors: Iterable[float],
    second_errors: Iterable[float],
    groups_ratio: float = 1.0,
    alternative: str = "two-sided",
    stabilizing_method: str = "asin",
) -> pd.DataFrame:
    """
    Get table for designing samples size for experiment.

    Parameters
    ----------
    dataframe: pd.DataFrame
        Table for designing the experiment
    column: Column name type
        Column, containg metric for designing
    effects: Iterable of floats
        List of effects which we want to catch.
        e.x.: [1.01, 1.02, 1.05]
    first_errors: Iterable of floats
        1st and 2nd type errors.
        e.x.: [0.01, 0.05, 0.1]
    second_errors: Iterable of floats
        1st and 2nd type errors.
        e.x.: [0.01, 0.05, 0.1]
    groups_ratio : float, default: ``1.0``
        Ratio between two groups.
    alternative : str, default: ``"two-sided"``
        Alternative hypothesis, can be ``"two-sided"``, ``"greater"``
        or ``"less"``.
        ``"greater"`` - if effect is positive.
        ``"less"`` - if effect is negative.
    stabilizing_method : str, default: ``"asin"``
        Effect trasformation. Can be ``"asin"`` and ``"norm"``.
        For non-binary metrics: only ``"norm"`` is accceptable.
        For binary metrics: ``"norm"`` and ``"asin"``, but ``"asin"``
        is more robust and accurate.

    Returns
    -------
    df_results : pd.DataFrame
        Table with minimal sample sizes for each effect and error from input data.
    """
    target_type = check_target_type(dataframe, column)
    mean, std = get_table_stats(dataframe, column)
    return get_table_sample_size(
        mean, std, effects, first_errors, second_errors, target_type, groups_ratio, alternative, stabilizing_method
    )


def get_minimal_effects_table(
    mean: float,
    std: float,
    sample_sizes: Iterable[int],
    first_errors: Iterable[float],
    second_errors: Iterable[float],
    as_numeric: bool = False,
    target_type: str = "non-binary",
    groups_ratio: float = 1.0,
    alternative: str = "two-sided",
    stabilizing_method: str = "asin",
) -> pd.DataFrame:
    """
    Create table of effects for different sample sizes and errors.

    Parameters
    ----------
    mean : float
        Sample mean
    std : float
        Sample standard deviation.
    sample_sizes : List
        List of sample sizes which we want to check.
        e.x.: [100, 200, 1000]
    first_errors : List
        1st and 2nd type errors.
        e.x.: [0.01, 0.05, 0.1]
    second_errors : List
        1st and 2nd type errors.
        e.x.: [0.01, 0.05, 0.1]
    as_numeric : bool, default False
        Whether to return a number or a string with percentages
    target_type : str, default: ``non-binary``
        Type of target metric: binary or non-binary
    groups_ratio : float, default: ``1.0``
        Ratio between two groups.
    alternative : str, default: ``"two-sided"``
        Alternative hypothesis, can be ``"two-sided"``, ``"greater"``
        or ``"less"``.
        ``"greater"`` - if effect is positive.
        ``"less"`` - if effect is negative.
    stabilizing_method : str, default: ``"asin"``
        Effect trasformation. Can be ``"asin"`` and ``"norm"``.
        For non-binary metrics: only ``"norm"`` is accceptable.
        For binary metrics: ``"norm"`` and ``"asin"``, but ``"asin"``
        is more robust and accurate.

    Returns
    -------
    df_results : pd.DataFrame
        Table with minimal effects for each sample size and error from input data.
    """
    multiindex = pd.MultiIndex.from_tuples([(size,) for size in sample_sizes], names=[GROUP_SIZE_COL_NAME])
    multicols = pd.MultiIndex.from_tuples(
        [(f"({err_one}; {err_two})",) for err_one in first_errors for err_two in second_errors],
        names=[STAT_ERRORS_COL_NAME],
    )
    df_results = pd.DataFrame(index=multiindex, columns=multicols)
    for sample_size in sample_sizes:
        for first_err in first_errors:
            for second_err in second_errors:
                err = f"({first_err}; {second_err})"
                effect = get_minimal_determinable_effect(
                    mean=mean,
                    std=std,
                    sample_size=sample_size,
                    alpha=first_err,
                    beta=second_err,
                    target_type=target_type,
                    groups_ratio=groups_ratio,
                    alternative=alternative,
                    stabilizing_method=stabilizing_method,
                )
                if as_numeric:
                    df_results.loc[(sample_size,), (err,)] = round(effect, ROUND_DIGITS_TABLE) + 1
                else:
                    df_results.loc[(sample_size,), (err,)] = str(np.round(effect * 100, ROUND_DIGITS_PERCENT)) + "%"
    df_results.index = pd.MultiIndex(
        levels=[sample_sizes],
        codes=[np.arange(len(sample_sizes))],
        names=[GROUP_SIZE_COL_NAME],
    )
    return df_results


def design_effect(
    dataframe: pd.DataFrame,
    column: types.ColumnNameType,
    sample_sizes: Iterable[int],
    first_errors: Iterable[float],
    second_errors: Iterable[float],
    as_numeric: bool = False,
    groups_ratio: float = 1.0,
    alternative: str = "two-sided",
    stabilizing_method: str = "asin",
) -> pd.DataFrame:
    """
    Create table of effects for different sample sizes and errors.

    Parameters
    ----------
    dataframe : pandas Data Frame
        Table for data to be designed
    column : Column name type
        Column of metric to be designed
    sample_sizes : List
        List of sample sizes which we want to check.
        e.x.: [100, 200, 1000]
    first_errors : List
        1st and 2nd type errors.
        e.x.: [0.01, 0.05, 0.1]
    second_errors : List
        1st and 2nd type errors.
        e.x.: [0.01, 0.05, 0.1]
    as_numeric : bool, default False
        Whether to return a number or a string with percentages
    groups_ratio : float, default: ``1.0``
        Ratio between two groups.
    alternative : str, default: ``"two-sided"``
        Alternative hypothesis, can be ``"two-sided"``, ``"greater"``
        or ``"less"``.
        ``"greater"`` - if effect is positive.
        ``"less"`` - if effect is negative.
    stabilizing_method : str, default: ``"asin"``
        Effect trasformation. Can be ``"asin"`` and ``"norm"``.
        For non-binary metrics: only ``"norm"`` is accceptable.
        For binary metrics: ``"norm"`` and ``"asin"``, but ``"asin"``
        is more robust and accurate.

    Returns
    -------
    df_results : pd.DataFrame
        Table with minimal effects for each sample size and error from input data.
    """
    target_type = check_target_type(dataframe, column)
    mean, std = get_table_stats(dataframe, column)
    return get_minimal_effects_table(
        mean,
        std,
        sample_sizes,
        first_errors,
        second_errors,
        as_numeric,
        target_type,
        groups_ratio,
        alternative,
        stabilizing_method,
    )


def get_power_table(
    mean: float,
    std: float,
    sample_sizes: Iterable[int],
    effects: Iterable[float],
    first_errors: Iterable[float] = (FIRST_TYPE_ERROR,),
    as_numeric: bool = False,
    target_type: str = "non-binary",
    groups_ratio: float = 1.0,
    alternative: str = "two-sided",
    stabilizing_method: str = "asin",
) -> pd.DataFrame:
    """
    Create table of power for different sample sizes and effects.

    Parameters
    ----------
    mean : float
        Sample mean
    std : float
        Sample standard deviation.
    sample_sizes : Iterable[int]
        List of sample sizes which we want to check.
        e.x.: [100, 200, 1000]
    effects : Iterable[float]
        Iterable object with
        e.x.: [1.01, 1.02, 1.05]
    first_errors : Iterable, default: ``(0.05,)``
        1st and 2nd type errors.
        e.x.: [0.01, 0.05, 0.1]
    as_numeric : bool, default False
        Whether to return a number or a string with percentages
    target_type : str, default: ``non-binary``
        Type of target metric: binary or non-binary
    groups_ratio : float, default: ``1.0``
        Ratio between two groups.
    alternative : str, default: ``"two-sided"``
        Alternative hypothesis, can be ``"two-sided"``, ``"greater"``
        or ``"less"``.
        ``"greater"`` - if effect is positive.
        ``"less"`` - if effect is negative.
    stabilizing_method : str, default: ``"asin"``
        Effect trasformation. Can be ``"asin"`` and ``"norm"``.
        For non-binary metrics: only ``"norm"`` is accceptable.
        For binary metrics: ``"norm"`` and ``"asin"``, but ``"asin"``
        is more robust and accurate.

    Returns
    -------
    df_results : pd.DataFrame
        Table with  sample sizes for each effect and error from input data.
    """
    effects_str = [str(round((effect - 1) * 100, ROUND_DIGITS_PERCENT)) + "%" for effect in effects]
    multiindex = pd.MultiIndex.from_tuples(
        [(first_error, effect_str) for first_error in first_errors for effect_str in effects_str],
        names=[FIRST_TYPE_ERROR_COL_NAME, EFFECT_COL_NAME],
    )
    powers: List[np.ndarray] = []
    for first_err in first_errors:
        for effect in effects:
            power: np.ndarray = get_power(
                mean=mean,
                std=std,
                sample_size=np.array(sample_sizes),
                effect=effect,
                alpha=first_err,
                target_type=target_type,
                groups_ratio=groups_ratio,
                alternative=alternative,
                stabilizing_method=stabilizing_method,
            )
            if as_numeric:
                power = [np.round(p, ROUND_DIGITS_TABLE) for p in power]
            else:
                power = [str(np.round(p * 100, ROUND_DIGITS_PERCENT)) + "%" for p in power]
            powers.append(power)
    df_results = pd.DataFrame(
        np.vstack(powers),
        columns=sample_sizes,
        index=multiindex,
    )
    df_results.index.name = "Errors and Effects"
    df_results.columns.name = "Group sizes"
    return df_results


def design_power(
    dataframe: pd.DataFrame,
    column: types.ColumnNameType,
    sample_sizes: Iterable[int],
    effects: Iterable[float],
    first_errors: Iterable[float] = (FIRST_TYPE_ERROR,),
    as_numeric: bool = False,
    groups_ratio: float = 1.0,
    alternative: str = "two-sided",
    stabilizing_method: str = "asin",
) -> pd.DataFrame:
    """
    Create table of power for different sample sizes and effects.

    Parameters
    ----------
    dataframe : pd.DataFrame
        Table for data to be designed
    column : Column name type
        Sample standard deviation.
    sample_sizes : Iterable[int]
        List of sample sizes which we want to check.
        e.x.: [100, 200, 1000]
    effects : Iterable[float]
        Iterable object with
        e.x.: [1.01, 1.02, 1.05]
    first_errors : Iterable[float], default: ``(0.05,)``
        1st and 2nd type errors.
        e.x.: [0.01, 0.05, 0.1]
    as_numeric : bool, default False
        Whether to return a number or a string with percentages
    groups_ratio : float, default: ``1.0``
        Ratio between two groups.
    alternative : str, default: ``"two-sided"``
        Alternative hypothesis, can be ``"two-sided"``, ``"greater"``
        or ``"less"``.
        ``"greater"`` - if effect is positive.
        ``"less"`` - if effect is negative.
    stabilizing_method : str, default: ``"asin"``
        Effect trasformation. Can be ``"asin"`` and ``"norm"``.
        For non-binary metrics: only ``"norm"`` is accceptable.
        For binary metrics: ``"norm"`` and ``"asin"``, but ``"asin"``
        is more robust and accurate.

    Returns
    -------
    df_results : pd.DataFrame
        Table with power for each effect and samples from input data.
    """
    target_type = check_target_type(dataframe, column)
    mean, std = get_table_stats(dataframe, column)
    return get_power_table(
        mean,
        std,
        sample_sizes,
        effects,
        first_errors,
        as_numeric,
        target_type,
        groups_ratio,
        alternative,
        stabilizing_method,
    )


def get_ttest_info_from_stats(
    var_a: float, var_b: float, n_obs_a: int, n_obs_b: int, alpha: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Returns quantiles and standard deviation of Ttest criterion statistic
    """
    compound_se: float = np.sqrt(var_a / n_obs_a + var_b / n_obs_b)
    denominator: float = (var_a / n_obs_a) ** 2 / (n_obs_a - 1) + (var_b / n_obs_b) ** 2 / (n_obs_b - 1)
    dim: float = compound_se**2 / denominator
    quantiles: np.ndarray = sps.t.ppf(1 - alpha / 2, df=dim)
    return quantiles, compound_se


def get_ttest_info(group_a: np.ndarray, group_b: np.ndarray, alpha: np.ndarray) -> Tuple[np.ndarray, float]:
    """
    Compute standart error and quatiles for ttest (Welch)
    https://en.wikipedia.org/wiki/Student%27s_t-test

    Arguments
    ---------
    group_a : np.ndarray
        Group A values
    group_b : np.ndarray
        Group B values
    alpha : np.ndarray
        First type error values

    Returns:
    -------
    quantiles, compound_se : np.ndarray, float
        Quantiles of T_dim corresponding each alpha
        compound_se - Standart error used in t-test
    """
    variance_group_a: float = group_a.var(ddof=1)
    variance_group_b: float = group_b.var(ddof=1)
    return get_ttest_info_from_stats(variance_group_a, variance_group_b, len(group_a), len(group_b), alpha)


def apply_delta_method_by_stats(
    size: int,
    mean_group_a: float,
    var_group_a: float,
    mean_group_b: float,
    var_group_b: float,
    cov_groups: float = 0,
    transformation: str = "fraction",
    alpha: np.ndarray = np.array([FIRST_TYPE_ERROR]),
    alternative: str = "two-sided",
) -> Tuple[types.ManyIntervalType, float]:
    """
    Computation of pvalue and confidence intervals for each I type error bound (alpha)
    using Delta-method by statistics.

    Arguments
    ---------
    size: int
        Size of both groups
    mean_group_a: float
        Mean of metrics from group A
    var_group_a: float
        Consistent estimation of variation (ddof = 0) of metrics from group A
    mean_group_b: float
        Mean of metrics from group B
    var_group_b: float
        Consistent estimation of variation (ddof = 0) of metrics from group B
    cov_groups: float, default ``0``
        Covariation between groups for dependent samples
    transformation : str, default: ``fraction``
        Continuous transformation of random variable
    alpha : np.ndarray, default: ``np.array([0.05])``
        Lists of I type errors bounds
    alternative : str, default: ``two-sided``
        Alternative for static criteria - two-sided, less, greater
        Less means, that mean in first group less, than mean in second group.
    Returns
    -------
    (intervals, pvalue) : Tuple[types.ManyIntervalType, float]
        Confidence intervals with given sufficient level
        And
        Pvalue for corresponding criterion
            H0: mean(A) / mean(B) = 1
                vs
            H1: mean(A) / mean(B) <> 1
    """
    admissible_transformations: List[str] = ["fraction"]
    if transformation in admissible_transformations:
        conf_intervals = pvalue_pkg.calculate_intervals_by_delta_method(
            mean_group_a, mean_group_b, var_group_a, var_group_b, cov_groups, size, transformation, alpha, alternative
        )
        pvalue = pvalue_pkg.calculate_pvalue_by_delta_method(
            mean_group_a, mean_group_b, var_group_a, var_group_b, cov_groups, size, transformation, alternative
        )
    else:
        raise ValueError(f'Choose method from {", ".join(admissible_transformations)}, got {transformation}')
    return conf_intervals, pvalue


def apply_delta_method(
    group_a: np.ndarray,
    group_b: np.ndarray,
    transformation: str,
    alpha: np.ndarray = np.array([FIRST_TYPE_ERROR]),
    dependent: bool = False,
    alternative: str = "two-sided",
) -> Tuple[types.ManyIntervalType, float]:
    """
    Computation of pvalue and confidence intervals for each I type error bound
    (alpha) using Delta-method.

    Parameters
    ----------
    group_a : np.ndarray
        Metrics from group A
    group_b : np.ndarray
        Metrics from group B
    transformation : str
        Continuous transformation of random variable
    alpha : np.ndarray, default: ``np.array([0.05])``
        Lists of I type errors bounds
    alternative : str, default: ``two-sided``
        Alternative for static criteria - two-sided, less, greater
        Less means, that mean in first group less, than mean in second group.

    Returns
    -------
    (intervals, pvalue) : Tuple[types.ManyIntervalType, float]
        Confidence intervals with given sufficient level
        And
        Pvalue for corresponding criterion
            H0: mean(A) / mean(B) = 1
                vs
            H1: mean(A) / mean(B) <> 1
    """
    mean_size: int = (group_a.shape[0] + group_b.shape[0]) // 2
    mean_a, std_group_a = get_stats(group_a, ddof=0)
    mean_b, std_group_b = get_stats(group_b, ddof=0)
    covariance_ab: float = np.cov(group_a, group_b)[0][1] if dependent else 0
    return apply_delta_method_by_stats(
        mean_size,
        mean_a,
        std_group_a**2,
        mean_b,
        std_group_b**2,
        covariance_ab,
        transformation,
        alpha,
        alternative=alternative,
    )


def ttest_1samp_from_stats(
    mean: float, std: float, n_obs: int, alternative: str = Alternatives.ts.value
) -> Tuple[float, float]:
    """
    Implementation of ttest_1samp for stats, not from observations

    Parameters
    ----------
    mean: float
        Mean of samples
    std: float
        Standart deviation (consistent estimation)
    n_obs: int
        Amount of observations
    alternative: str
        One of two-sided, less, greater

    Returns
    -------
    (statistic, pvalue): Tuple[float, float]
    Statistic of criterion and pvalue
    """
    statistic: float = mean / std * np.sqrt(n_obs)

    if alternative == Alternatives.gr.value:
        pvalue: float = sps.t.sf(statistic, df=n_obs - 1)
    elif alternative == Alternatives.less.value:
        pvalue: float = sps.t.cdf(statistic, df=n_obs - 1)
    elif alternative == Alternatives.ts.value:
        pvalue: float = sps.t.cdf(-np.abs(statistic), df=n_obs - 1) + sps.t.sf(np.abs(statistic), df=n_obs - 1)
    else:
        Alternatives.raise_if_value_incorrect_enum(alternative)
    return statistic, pvalue



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""
Module contains class for data aggregation during a preprocessing task.
"""
import copy
from typing import Any, Dict, Optional, Union

import pandas as pd

from ambrosia import types
from ambrosia.tools.ab_abstract_component import AbstractFittableTransformer
from ambrosia.tools.back_tools import wrap_cols


class AggregatePreprocessor(AbstractFittableTransformer):
    """
    Preprocessing class for data aggregation.

    Can group data by multiple columns and aggregate it using methods
    for real and categorial features.

    Parameters
    ----------
    categorial_method : types.MethodType, default: ``"mode"``
        Aggregation method for categorial variables that
        will become as a default behavior.
    real_method : types.MethodType, default: ``"sum"``
        Aggregation method for real variables that
        will become as a default behavior.

    Attributes
    ----------
    categorial_method : types.MethodType
        Default aggregation method for categorial variables.
    real_method : types.MethodType
        Default aggregation method for real variables.
    groupby_columns : types.ColumnNamesType
        Columns which were used for groupping in the last aggregation.
        Gets value after fitting the class instance.
    agg_params : Dict
        Dictionary with aggregation rules which was used in the last
        aggregation.
        Gets value after fitting the class instance.
    """

    @staticmethod
    def __mode_calculation(values: pd.Series) -> Any:
        """
        Mode function for aggregation.
        """
        return values.value_counts().index[0]

    @staticmethod
    def __simple_agg(values: pd.Series) -> Any:
        """
        Simple aggregation, just picks the first element.
        """
        return values.iloc[0]

    @staticmethod
    def __transform_agg_param(aggregation_method: types.MethodType) -> types.MethodType:
        """
        Invoke an aggregation callable function by given string alias.
        """
        if aggregation_method == "mode":
            return AggregatePreprocessor.__mode_calculation
        if aggregation_method == "simple":
            return AggregatePreprocessor.__simple_agg
        return aggregation_method

    @staticmethod
    def __transform_params(dataframe: pd.DataFrame, aggregation_params: Dict) -> Dict:
        """
        Iteratively apply transformations specified by aggragation parameters.
        """
        agg_params = copy.deepcopy(aggregation_params)
        for column, method in agg_params.items():
            if column not in dataframe.columns:
                raise ValueError(f"{column} does not exist in the dataframe!")
            agg_params[column] = AggregatePreprocessor.__transform_agg_param(method)
        return agg_params

    def __init__(self, categorial_method: types.MethodType = "mode", real_method: types.MethodType = "sum"):
        self.categorial_method = categorial_method
        self.real_method = real_method
        self.agg_params = None
        self.groupby_columns = None
        super().__init__()

    def __real_case_step(
        self,
        agg_params: Optional[Dict] = None,
        real_cols: Optional[types.ColumnNamesType] = None,
    ) -> None:
        """
        A private method containing aggregation parameters filling logic
        for real metrics.
        """
        real_cols = wrap_cols(real_cols)
        for real_feature in real_cols:
            agg_params[real_feature] = self.real_method

    def __categorial_case_step(
        self,
        agg_params: Optional[Dict] = None,
        categorial_cols: Optional[types.ColumnNamesType] = None,
    ) -> None:
        """
        A private method containing aggregation parameters filling logic
        for categorial metrics.
        """
        categorial_cols = wrap_cols(categorial_cols)
        for categorial_feature in categorial_cols:
            agg_params[categorial_feature] = self.categorial_method

    def __empty_args_step(
        self,
        agg_params: Optional[Dict] = None,
        real_cols: Optional[types.ColumnNamesType] = None,
        categorial_cols: Optional[types.ColumnNamesType] = None,
    ) -> None:
        """
        A private method containing aggregation parameters filling logic
        if no aggregation parameters passed.
        """
        if real_cols is not None:
            self.__real_case_step(agg_params, real_cols)
        if categorial_cols is not None:
            self.__categorial_case_step(agg_params, categorial_cols)

    def get_params_dict(self) -> Dict:
        """
        Returns dictionary with parameters of the last run() or transform() call.
        """
        self._check_fitted()
        return {"aggregation_params": self.agg_params, "groupby_columns": self.groupby_columns}

    def load_params_dict(self, params: Dict) -> None:
        """
        Load prefitted parameters form a dictionary.

        Parameters
        ----------
        params : Dict
            Dictionary with prefitted params.
        """
        if "groupby_columns" in params:
            self.groupby_columns = params["groupby_columns"]
        else:
            raise TypeError(f"params argument must contain: {'column_names'}")
        if "aggregation_params" in params:
            self.agg_params = params["aggregation_params"]
        else:
            raise TypeError(f"params argument must contain: {'aggregation_params'}")
        self.fitted = True

    def fit(
        self,
        dataframe: pd.DataFrame,
        groupby_columns: types.ColumnNamesType,
        agg_params: Optional[Dict] = None,
        real_cols: Optional[types.ColumnNamesType] = None,
        categorial_cols: Optional[types.ColumnNamesType] = None,
    ) -> pd.DataFrame:
        """
        Fit preprocessor with parameters of aggregation.

        Aggregation will be performed using passed dictionary with
        defined aggregation conditions for each columns of interest,
        or lists of columns with default class aggregation behavior.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Table with selected columns.
        groupby_columns : types.ColumnNamesType
            Columns for GROUP BY.
        agg_params : Dict, optional
            Dictionary with aggregation parameters.
        real_cols : types.ColumnNamesType, optional
            Columns with real metrics.
            Overriden by ``agg_params`` parameter and could be passed if
            expected default aggregation behavior.
        categorial_cols : types.ColumnNamesType, optional
            Columns with categorial metrics
            Overriden by ``agg_params`` parameter and could be passed if
            expected default aggregation behavior.

        Returns
        -------
        self : object
            Instance object.
        """
        if agg_params is None and real_cols is None and categorial_cols is None:
            raise ValueError("Set agg_params or pass real_cols and categorial_cols")
        if agg_params is None:
            agg_params = {}
            self.__empty_args_step(agg_params, real_cols, categorial_cols)
        self._check_cols(dataframe, agg_params.keys())
        self.groupby_columns = groupby_columns
        self.agg_params = copy.deepcopy(agg_params)
        self.fitted = True
        return self

    def transform(
        self,
        dataframe: pd.DataFrame,
    ) -> pd.DataFrame:
        """
        Apply table transformation by its aggregation with prefitted
        parameters.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Table to aggregate.

        Returns
        -------
        agg_table : pd.DataFrame
            Aggregated table.
        """
        self._check_fitted()
        self._check_cols(dataframe, self.agg_params.keys())
        agg_params = AggregatePreprocessor.__transform_params(dataframe, self.agg_params)
        return dataframe.groupby(self.groupby_columns, as_index=False).agg(agg_params)

    def fit_transform(
        self,
        dataframe: pd.DataFrame,
        groupby_columns: types.ColumnNamesType,
        agg_params: Optional[Dict] = None,
        real_cols: Optional[types.ColumnNamesType] = None,
        categorial_cols: Optional[types.ColumnNamesType] = None,
    ) -> pd.DataFrame:
        """
        Fit preprocessor parameters using given dataframe and aggregate it.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Table to aggregate.
        groupby_columns : types.ColumnNamesType
            Columns for GROUP BY.
        agg_params : Dict, optional
            Dictionary with aggregation parameters.
        real_cols : types.ColumnNamesType, optional
            Columns with real metrics.
            Overriden by ``agg_params`` parameter and could be passed if
            expected default aggregation behavior.
        categorial_cols : types.ColumnNamesType, optional
            Columns with categorial metrics
            Overriden by ``agg_params`` parameter and could be passed if
            expected default aggregation behavior.

        Returns
        -------
        agg_table : pd.DataFrame
            Aggregated table.
        """
        self.fit(dataframe, groupby_columns, agg_params, real_cols, categorial_cols)
        return self.transform(dataframe)



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, Iterable, Optional

import ambrosia.tools.ab_abstract_component as ab_abstract
from ambrosia import types
from ambrosia.tools.import_tools import spark_installed

if spark_installed():
    import pyspark.sql.functions as spark_funcs
    from pyspark.sql import Window


EMPTY_VALUE: int = 0
STRAT_GROUPS: str = "__ambrosia_strat"


class Stratification(ab_abstract.StratificationUtil):
    """
    Stratification implementation for spark tables
    https://en.wikipedia.org/wiki/Stratified_sampling
    """

    def fit(self, dataframe: types.SparkDataFrame, columns: Optional[Iterable[types.ColumnNameType]] = None):
        if columns is None:
            self.strats = {ab_abstract.EmptyStratValue.NO_STRATIFICATION: dataframe}
            return

        window = Window.orderBy(*columns).partitionBy(spark_funcs.lit(EMPTY_VALUE))
        with_groups = dataframe.withColumn(STRAT_GROUPS, spark_funcs.dense_rank().over(window))
        amount_of_strats: int = with_groups.select(spark_funcs.max(STRAT_GROUPS)).collect()[0][0]

        self.strats: Dict[int, types.SparkDataFrame] = {}
        for strat_value in range(1, amount_of_strats + 1):
            strat_table = with_groups.where(spark_funcs.col(STRAT_GROUPS) == strat_value)
            self.strats[strat_value] = strat_table.drop(STRAT_GROUPS)

    def strat_sizes(self) -> Dict[int, int]:
        """
        Returns size of each stratification group
        """
        self._check_fit()
        sizes: Dict[Any, int] = {}
        for value, table in self.strats.items():
            sizes[value] = table.count()
        return sizes



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


# Do we need exceptions ?
# If package is not installed we can't pass such object to Solvers (Designer, ... )


import typing as tp
from importlib.metadata import distributions


class NotInstalledPackage(Exception):
    default_message: str = "This package is not installed"

    def __init_subclass__(cls, default_message: str) -> None:
        cls.default_message = default_message
        return super().__init_subclass__()

    def __init__(self, *args, **kwargs):
        if args:
            super().__init__(*args, **kwargs)
        else:
            super().__init__(self.default_message, **kwargs)


class PysparkNotInstalled(NotInstalledPackage, default_message="Install pyspark or ambrosia[spark]"):
    pass


def get_installed_package_names() -> tp.List[str]:
    return [dist.metadata["Name"].lower() for dist in distributions()]


def check_package(package_name: str) -> bool:
    return package_name in get_installed_package_names()


def spark_installed() -> bool:
    return check_package("pyspark")


def check_spark_installed() -> tp.NoReturn:
    if not spark_installed():
        raise PysparkNotInstalled()



import enum
import typing as tp

import numpy as np
import pandas as pd

import ambrosia.spark_tools.stat_criteria as spark_crit_pkg
import ambrosia.tools.stat_criteria as criteria_pkg
from ambrosia import types
from ambrosia.tools.ab_abstract_component import StatCriterion, choose_on_table
from ambrosia.tools.import_tools import spark_installed

# Avoid errors with not installed spark

if spark_installed():
    import pyspark.sql.functions as spark_funcs


def filter_spark_and_make_groups(
    dataframe: types.SparkDataFrame,
    df_mapping: types.GroupsInfoType,
    column_groups: types.ColumnNameType,
    group_labels: types.GroupLabelsType,
    id_column: types.ColumnNameType,
) -> types.TwoSamplesType:
    if dataframe is None:
        return None
    if df_mapping is not None:
        raise NotImplementedError("For spark tables df_mapping can't be used. Use column_groups instead")
    group_labels = dataframe.select(column_groups).distinct().collect()
    group_labels = sorted([label[column_groups] for label in group_labels])
    experiment_results: types.ExperimentResults = {
        label: dataframe.where(spark_funcs.col(column_groups) == label) for label in group_labels
    }
    return experiment_results


class PandasCriteria(enum.Enum):
    ttest: StatCriterion = criteria_pkg.TtestIndCriterion
    ttest_rel: StatCriterion = criteria_pkg.TtestRelCriterion
    mw: StatCriterion = criteria_pkg.MannWhitneyCriterion
    wilcoxon: StatCriterion = criteria_pkg.WilcoxonCriterion


class SparkCriteria(enum.Enum):
    ttest: StatCriterion = spark_crit_pkg.TtestIndCriterionSpark
    ttest_rel: StatCriterion = None  # spark_crit_pkg.TtestRelativeCriterionSpark it's in development now
    mw: StatCriterion = None
    wilcoxon: StatCriterion = None


class TheoreticalTesterHandler:
    def __init__(
        self, group_a, group_b, column: str, alpha: np.ndarray, effect_type: str, criterion: StatCriterion, **kwargs
    ):
        self.group_a = group_a
        self.group_b = group_b
        self.column = column
        self.alpha = alpha
        self.effect_type = effect_type
        self.criterion = criterion
        self.kwargs = kwargs

    def _correct_criterion(self, criterion: tp.Any) -> bool:
        return isinstance(criterion, StatCriterion)

    def _raise_correct_criterion(self, criterion: tp.Any) -> None:
        if not self._correct_criterion(criterion):
            raise TypeError("Criterion must be inherited from StatCriterion")

    def get_criterion(self, criterion: str, data_example: types.SparkOrPandas):
        if not isinstance(criterion, str):
            return criterion
        CriteriaEnum = choose_on_table([PandasCriteria, SparkCriteria], data_example)
        criterion = CriteriaEnum[criterion].value
        if criterion is None:
            raise NotImplementedError("This criterion will be implemented later")
        return criterion()

    def _set_kwargs(self):
        if isinstance(self.group_a, pd.DataFrame):
            self.group_a = self.group_a[self.column].values
            self.group_b = self.group_b[self.column].values
        elif isinstance(self.group_a, types.SparkDataFrame):
            self.kwargs["column"] = self.column
        self.kwargs["alpha"] = self.alpha
        self.kwargs["effect_type"] = self.effect_type
        self.kwargs["group_a"] = self.group_a
        self.kwargs["group_b"] = self.group_b

    def solve(self) -> types._SubResultType:
        criterion: tp.Union[str, StatCriterion] = self.criterion if self.criterion is not None else "ttest"
        criterion = self.get_criterion(criterion, self.group_a)
        self._raise_correct_criterion(criterion)
        self._set_kwargs()
        return criterion.get_results(**self.kwargs)



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from itertools import product
from typing import Any, Callable, Dict, Iterable, List, Tuple

import pandas as pd
from tqdm.auto import tqdm

ROUND_DIGITS_TABLE: int = 3
ROUND_DIGITS_PERCENT: int = 1


class Selector:
    """
    Variate params for designing experiment.
    """

    def __init__(
        self, solver: Callable, selecting_params: Dict[str, Iterable], n_jobs: int = 1, use_tqdm: bool = True, **kwargs
    ) -> None:
        self.solver = solver
        self.sel_params = selecting_params
        self.kwargs = kwargs
        self.n_jobs = n_jobs
        self.tqdm = use_tqdm

    def set_params(self, params: Tuple[Any, ...]) -> None:
        for param_name, param_value in zip(self.sel_params.keys(), params):
            self.kwargs[param_name] = param_value

    def iterate_params(self) -> Tuple[Tuple[Any, ...], List[Any]]:
        parameters: Tuple = tuple(product(*list(self.sel_params.values())))
        result = []
        iter_set = parameters if not self.tqdm else tqdm(parameters)
        for params in iter_set:
            self.set_params(params)
            result.append(self.solver(**self.kwargs))
        return parameters, result

    @staticmethod
    def handle_numeric(report: pd.DataFrame, as_numeric: bool) -> None:
        if not as_numeric:
            report["effect"] = (round((report["effect"] - 1) * 100, ROUND_DIGITS_PERCENT)).astype(str) + "%"
        report["errors"] = tuple(zip(report["alpha"], report["beta"]))

    def get_table_size(self, as_numeric: bool = False) -> pd.DataFrame:
        parameters, group_sizes_list = self.iterate_params()
        report = pd.DataFrame(list(parameters), columns=["effect", "alpha", "beta"]).join(
            pd.DataFrame(list(group_sizes_list), columns=["sample_sizes"])
        )
        self.handle_numeric(report, as_numeric)
        report = report.pivot(index="effect", columns="errors", values="sample_sizes")
        report = report.sort_values(report.columns[0])
        return report

    def get_table_effect(self, as_numeric: bool = False) -> pd.DataFrame:
        parameters, effects = self.iterate_params()
        report = pd.DataFrame(list(parameters), columns=["group_sizes", "alpha", "beta"]).join(
            pd.DataFrame(list(effects), columns=["effect"])
        )
        self.handle_numeric(report, as_numeric)
        report = report.pivot(index="group_sizes", columns="errors", values="effect")
        report = report.sort_index()
        return report



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import Dict, List

import numpy as np
import pandas as pd

from ambrosia import types

NAME: str = "ambrosia_LOGGER"
PREFIX: str = "ambrosia LOGGER"


def info_log(message: str):
    logger = logging.Logger(NAME)
    logger.warning(f"{PREFIX}: {message}")


class RobustLogger:
    """
    Temporary class with methods for calculating and logging changes
    in the characteristics of metric distributions during the preprocessing.
    """

    @staticmethod
    def verbose(prev_stats: Dict[str, float], new_stats: Dict[str, float], name: str) -> None:
        """
        Verbose transormations to os.stdout.
        """
        for metric in prev_stats.keys():
            prev: float = prev_stats[metric]
            new: float = new_stats[metric]
            info_log(f"Change {metric} {name}: {prev:.4f} ===> {new:.4f}")

    @staticmethod
    def verbose_list(
        prev_stats: List[Dict[str, float]],
        new_stats: List[Dict[str, float]],
        names: types.ColumnNamesType,
    ) -> None:
        """
        Verbose iteratively.
        """
        for name, stat_1, stat_2 in zip(names, prev_stats, new_stats):
            info_log("\n")
            RobustLogger.verbose(stat_1, stat_2, name)

    @staticmethod
    def __calculate_stats(values: np.ndarray) -> Dict[str, float]:
        return {
            "Mean": np.mean(values),
            "Variance": np.var(values),
            "IQR": np.quantile(values, 0.75) - np.quantile(values, 0.25),
            "Range": np.max(values) - np.min(values),
        }

    @staticmethod
    def get_stats(
        df: pd.DataFrame,
        names: types.ColumnNamesType,
    ) -> List[Dict[str, float]]:
        """
        Get metrics for all columns.
        """
        result: List[Dict[str, float]] = []
        for name in names:
            err_msg: str = f"Column name is not in data frame, coumn - {name}"
            assert name in df.columns, err_msg
            result.append(RobustLogger.__calculate_stats(df[name].values))
        return result



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, List

import numpy as np

# Try nmslib first, fallback to hnswlib if it fails (e.g., on macOS ARM)
_USE_HNSWLIB = False
try:
    import nmslib
except ImportError:
    _USE_HNSWLIB = True
    nmslib = None

# Always try to import hnswlib as fallback
try:
    import hnswlib
except ImportError:
    hnswlib = None


class NMTree:
    """
    KNN solution with nmslib (primary) or hnswlib (fallback for macOS ARM).
    """

    def __init__(self, points: np.ndarray, payload: np.ndarray, ef_search: int):
        self.__was = set()
        self.__payload = payload
        self.__use_hnswlib = _USE_HNSWLIB

        if not self.__use_hnswlib:
            try:
                self.__index = nmslib.init(space="l2")
                self.__index.addDataPointBatch(points)
                self.__index.createIndex()
                self.__index.setQueryTimeParams({"efSearch": ef_search})
            except Exception:
                # nmslib crashed (segfault on macOS ARM), try hnswlib
                self.__use_hnswlib = True

        if self.__use_hnswlib:
            if hnswlib is None:
                raise ImportError(
                    "nmslib failed and hnswlib is not installed. " "Please install hnswlib: pip install hnswlib"
                )
            dim = points.shape[1]
            self.__index = hnswlib.Index(space="l2", dim=dim)
            self.__index.init_index(max_elements=len(points), ef_construction=200, M=16)
            self.__index.add_items(points, np.arange(len(points)))
            self.__index.set_ef(ef_search)

    def query_batch(
        self,
        points: np.ndarray,
        payload: np.ndarray,
        out: List[List[Any]],
        group_size: int,
        amount: int = 1,
        threads: int = 1,
    ) -> None:
        """
        Write to out list indices for groups.

        Parameters
        ----------
        points : np.ndarray
            Batch of points for query
        payload : np.ndarray
            Some usable information
        out : List[List[Any]]
            Output list for indices
        group_size : int
            Sizes for groups
        amount : int, default: ``1``
            Amount groups exclude one
        threads : int, default: ``1``
            Amount of threads to be used
        """
        k = group_size * amount

        if self.__use_hnswlib:
            self.__index.set_num_threads(threads)
            labels, _ = self.__index.knn_query(points, k=k)
            indices = [row for row in labels]
        else:
            neighbours = self.__index.knnQueryBatch(points, k=k, num_threads=threads)
            indices: List[np.ndarray] = [x[0] for x in neighbours]

        for i, index in enumerate(indices):
            current_answer: List = []
            for j in index:
                if self.__payload[j] in self.__was:
                    continue
                self.__was.add(self.__payload[j])
                current_answer.append(self.__payload[j])
                if len(current_answer) == amount:
                    out.append(current_answer + [payload[i]])
                    break



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Methods for calculating test results for binary metrics.

Module contains functions that help to deal
with test results evaluation of binary metrics.

Mainly these functions are used in `Tester` core class.
"""
from typing import Tuple

import numpy as np
import scipy.stats as sps

import ambrosia.tools.bin_intervals as bin_pkg
import ambrosia.tools.pvalue_tools as pvalue_pkg
from ambrosia import types


def binary_absolute_result(
    group_a: types.GroupType, group_b: types.GroupType, alpha: np.ndarray, **kwargs
) -> types._SubResultType:
    """
    Function to calculate absolute test result for binary metrics.

    Parameters
    ----------
    group_a : types.GroupType
        Array containing the binary metrics for group A.
    group_b : types.GroupType
        Array containing the binary metrics for group B.
    alpha: np.ndarray
        Array of the first type errors.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    result_dict : _SubResultType
        Dict with absolute test result, computed for binary metrics.
    """
    success_a: int = group_a.sum()
    success_b: int = group_b.sum()
    trials_a: int = len(group_a)
    trials_b: int = len(group_b)
    pvalue: float = bin_pkg.BinomTwoSampleCI.calculate_pvalue(
        a_success=success_a, b_success=success_b, a_trials=trials_a, b_trials=trials_b, **kwargs
    )
    point_effect: float = np.mean(group_b) - np.mean(group_a)
    conf_intervals = bin_pkg.BinomTwoSampleCI.confidence_interval(
        a_success=success_a,
        b_success=success_b,
        a_trials=trials_a,
        b_trials=trials_b,
        confidence_level=1 - alpha,
        **kwargs,
    )
    conf_intervals = list(zip(conf_intervals[0], conf_intervals[1]))
    return {
        "first_type_error": alpha,
        "pvalue": pvalue,
        "effect": point_effect,
        "confidence_interval": conf_intervals,
    }


def binary_relative_confidence_interval(
    confidence_level: float, group_a: types.GroupType, group_b: types.GroupType, alternative: str = "two-sided"
) -> Tuple[float, float]:
    """
    Function for constructing a relative confidence intervals.

    Uses log-tranformation https://en.wikipedia.org/wiki/Relative_risk.

    Parameters
    ----------
    confidence_level : np.ndarray
        Array of nominal coverages of confidence intervals.
    group_a : types.GroupType
        Array containing the binary metrics for group A.
    group_b : types.GroupType
        Array containing the binary metrics for group B.

    Returns
    -------
    intervals_bounds : Tuple[np.ndarray, np.ndarray]
        Two arrays containing left and right bounds of binary
        relative confidence intervals.
    """
    pvalue_pkg.check_alternative(alternative)
    confidence_level = 1 - pvalue_pkg.corrected_alpha(1 - confidence_level, alternative)
    p_a: float = np.mean(group_a)
    p_b: float = np.mean(group_b)
    estimation_fraction: float = p_b / p_a
    a_size: int = len(group_a)
    b_size: int = len(group_b)
    se_rr: float = np.sqrt((1 - p_a) / (p_a * a_size) + (1 - p_b) / (p_b * b_size))
    quantiles: np.ndarray = sps.norm.ppf((1 + confidence_level) / 2)
    left: np.ndarray = np.log(estimation_fraction) - quantiles * se_rr
    right: np.ndarray = np.log(estimation_fraction) + quantiles * se_rr
    return pvalue_pkg.choose_from_bounds(np.exp(left) - 1, np.exp(right) - 1, alternative)


def binary_relative_result(
    group_a: types.GroupType, group_b: types.GroupType, alpha: np.ndarray, alternative: str = "two-sided"
) -> types._SubResultType:
    """
    Function calculates relative result for binary metrics
    Using log-tranformation https://en.wikipedia.org/wiki/Relative_risk
    """
    point_effect = np.mean(group_b) / np.mean(group_a) - 1
    function_interval = binary_relative_confidence_interval
    intervals = function_interval(1 - alpha, group_a, group_b, alternative=alternative)
    confidence_intervals = [(left, right) for left, right in zip(*intervals)]
    pvalue: float = pvalue_pkg.calculate_pvalue_by_interval(
        function_interval, 0, group_a=group_a, group_b=group_b, alternative=alternative
    )
    return {
        "first_type_error": alpha,
        "pvalue": pvalue,
        "effect": point_effect,
        "confidence_interval": confidence_intervals,
    }



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Handlers for different dataframe types.

Module containes functions that help to deal
with data of different type during the groups split task.

Mainly these functions are used in `Splitter` core class.
"""
from typing import List, Optional

import pandas as pd

import ambrosia.spark_tools.split_tools as split_spark
import ambrosia.tools.split_tools as split_pandas
from ambrosia import types
from ambrosia.tools.import_tools import spark_installed

# Avoid errors with not installed spark

if spark_installed():
    import pyspark.sql.functions as spark_funcs


AVAILABLE: List[str] = ["pandas", "spark"]
GROUPS_COLUMN: str = "group"
DATA: str = "dataframe"
THREADS: str = "threads"


def add_data_pandas(dataframe: pd.DataFrame, splitted_dataframe: pd.DataFrame, group_label: str) -> pd.DataFrame:
    """
    Add data to splitted dataframe for pandas tables.

    Parameters
    ----------
    dataframe : pd.DataFrame
        Source full dataframe.
    splitted_dataframe : pd.DataFrame
        Part of dataframe, which was splitted into two groups.
    group_label : str
        Group label which will be set.

    Returns
    -------
    customized_table : pd.DataFrame
        Table, where rest of table is added.
    """
    additional_table: pd.DataFrame = dataframe.loc[dataframe.index.delete(splitted_dataframe.index)].copy()
    additional_table[GROUPS_COLUMN] = group_label
    customized_table = pd.concat([splitted_dataframe, additional_table])
    return customized_table


def add_data_spark(
    dataframe: types.SparkDataFrame,
    splitted_dataframe: types.SparkDataFrame,
    group_label: str,
    id_column: types.ColumnNamesType,
) -> types.SparkDataFrame:
    """
    Add data to splitted dataframe for spark tables.

    Parameters
    ----------
    dataframe : SparkDataFrame
        Source full dataframe.
    splitted_dataframe: SparkDataFrame
        Part of dataframe, which was splitted into two groups.
    group_label : str
        Group label which will be set.
    id_column : ColumnNamesType
        Column with id's.

    Returns
    -------
    customized_table : SparkDataFrame
        Table, where rest of table added.
    """
    new_data = (
        dataframe.join(splitted_dataframe.select(id_column, GROUPS_COLUMN), on=id_column, how="left")
        .where(spark_funcs.col(GROUPS_COLUMN).isNull())
        .withColumn(GROUPS_COLUMN, spark_funcs.lit(group_label))
    )
    return splitted_dataframe.union(new_data)


def add_data_to_splitted(
    dataframe: types.PassedDataType,
    splitted_dataframe: types.SplitterResult,
    group_label: str,
    id_column: Optional[types.ColumnNamesType] = None,
) -> types.SplitterResult:
    """
    Add data to splitted groups.

    Parameters
    ----------
    dataframe : PassedDataType
        Source full dataframe.
    splitted_dataframe :  SplitterResult
        Part of dataframe, which was splitted into two groups.
    group_label : str
        Group label which will be set.
    id_columns : ColumnNamesType, optional
        Columns with id's for spark tables.

    Returns
    -------
    customized_table : SplitterResult
        Table, where rest of table added.
    """
    if isinstance(dataframe, pd.DataFrame):
        return add_data_pandas(dataframe, splitted_dataframe, group_label)
    elif isinstance(dataframe, types.SparkDataFrame):
        return add_data_spark(dataframe, splitted_dataframe, group_label, id_column)
    else:
        raise TypeError(f'Type of table must be one of {", ".join(AVAILABLE)}')


def handle_full_split(
    dataframe: types.PassedDataType,
    splitted_dataframe: types.SplitterResult,
    split_factor: float,
    id_column: Optional[types.ColumnNamesType] = None,
) -> types.SplitterResult:
    """
    Finish split dataframe according to split_factor.
    """
    if split_factor < 0.5:
        group_label = "B"
    else:
        group_label = "A"
    return add_data_to_splitted(dataframe, splitted_dataframe, group_label, id_column)


def data_shape(dataframe: types.PassedDataType) -> int:
    """
    Calculate table length size in cases of different tables.
    """
    if isinstance(dataframe, pd.DataFrame):
        return dataframe.shape[0]
    elif isinstance(dataframe, types.SparkDataFrame):
        return dataframe.count()
    else:
        raise TypeError(f'Type of table must be one of {", ".join(AVAILABLE)}')


def split_data_handler(**kwargs) -> types.SplitterResult:
    """
    Call split function according to table type.
    """
    if isinstance(kwargs[DATA], pd.DataFrame):
        return split_pandas.get_split(**kwargs)
    elif isinstance(kwargs[DATA], types.SparkDataFrame):
        return split_spark.get_split(**kwargs)
    else:
        raise TypeError(f'Type of table must be one of {", ".join(AVAILABLE)}')



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from abc import ABC, ABCMeta, abstractmethod
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Tuple, Union

import numpy as np
import pandas as pd
from yaml import YAMLObjectMetaclass

import ambrosia.tools.pvalue_tools as pvalue_pkg
from ambrosia import types
from ambrosia.tools import log

AVAILABLE: List[str] = ["pandas", "spark"]
DATA: str = "dataframe"


class ABMetaClass(ABCMeta, YAMLObjectMetaclass):
    """
    A metaclass to solve the problem 'metaclass conflict'.
    If you want to derive from ABToolAbstract set metaclass=ABMetaClass.
    """


class ABToolAbstract(ABC):
    _SavedArgumentType = Any
    _GivenArgument_Type = Any
    _PrepareArgumentsType = Dict[str, Tuple[_SavedArgumentType, _GivenArgument_Type]]
    _PRECISION_DIGITS = 4

    @abstractmethod
    def run(self):
        """
        Each derived class must implement this method.
        """

    @staticmethod
    def _prepare_arguments(_args: _PrepareArgumentsType) -> types._UsageArgumentsType:
        """
        Protected method for derived classes.
        Choose values for attributes (saved from __init__, passed to method run)
        If passed value not None, then it will be used, otherwise - saved value

        Parameters
        ----------
        _args : _PrepareArgumentsType
            Dictionary with
                keys - name of attribute of derived class
                values - (saved argument from constructor, given argument for run method)

        Returns
        -------
        chosen_args : _UsageArgumentsType
            Dictionary with
                keys - name of attribute
                values - chosen arg
        """
        chosen_args: types._UsageArgumentsType = {}
        for arg_name, [saved_argument, given_argument] in _args.items():
            exception_message: str = f"""Value for argument - {arg_name},
             was not set! Define it via setter method or pass as argument"""
            choice = given_argument if given_argument is not None else saved_argument
            if choice is None:
                raise ValueError(exception_message)
            chosen_args[arg_name] = choice
        return chosen_args


class AbstractFittableTransformer(ABC):
    """
    Abstract class for fittable transformer.

    Attributes
    ----------
    fitted : bool
        Indicator for the fit status of transformer.
    """

    def __init__(self):
        self.fitted: bool = False

    def _check_fitted(self) -> None:
        if not self.fitted:
            raise RuntimeError("Call fit method before!")

    def _check_cols(self, dataframe: pd.DataFrame, columns: types.ColumnNameType) -> None:
        for column in columns:
            if column not in dataframe:
                raise ValueError(f"Column {column} is not in Data Frame columns list")

    @abstractmethod
    def get_params_dict(self) -> Dict:
        """
        Returns a dictionary with params.
        """

    @abstractmethod
    def load_params_dict(self, params: Dict) -> None:
        """
        Load model parameters from the dictionary.

        Parameters
        ----------
        params : Dict
            Dictionary with params.
        """

    @abstractmethod
    def fit(self):
        """
        Fit class parameters on some data.
        """

    @abstractmethod
    def transform(self):
        """
        Transform data using fitted parameters.
        """

    @abstractmethod
    def fit_transform(self):
        """
        Fit class parameters on some data and transform it.
        """

    def store_params(self, store_path: Path) -> None:
        """
        Parameters
        ----------
        store_path : Path
            Path where parameters will be stored in a json format.
        """
        with open(store_path, "w+") as file:
            json.dump(self.get_params_dict(), file)

    def load_params(self, load_path: Path) -> None:
        """
        Parameters
        ----------
        load_path : Path
            Path to json file with parameters.
        """
        with open(load_path, "r+") as file:
            params = json.load(file)
            self.load_params_dict(params)


class AbstractVarianceReducer(AbstractFittableTransformer):
    """
    Abstract class for Variance Reduction.

    Parameters
    ----------
    verbose : bool, default: ``True``
        If ``True`` will print in sys.stdout the information
        about the variance reduction.

    Attributes
    ----------
    params : Dict
        Parameters of the VarianceReducer that will be updated after
        calling ``fit`` method.
        These parameters are sufficient for data frames transformations
        and are used when loading and saving methods of instance are called.
        By the default these params dictionary contains  ``'target_column'``
        and ``'transformed_name'`` keys.
    verbose : bool
        Verbose info flag.
    fitted : bool
        Indicator for the fit status of transformer.
    """

    EPSILON: float = 1e-5

    @abstractmethod
    def __call__(self, y: np.ndarray, X: np.ndarray) -> np.ndarray:  # pylint: disable=C0103
        """
        Returns transformed y values.

        Returns
        -------
        y_hat: np.ndarray
            y_hat = y - transformation(X) + MEAN(transformation(X))
        """

    def __init__(self, verbose: bool = True) -> None:
        self.params = {"target_column": None, "transformed_name": None}
        self.verbose: bool = verbose
        super().__init__()

    def _return_result(
        self, dataframe: pd.DataFrame, new_target: np.ndarray, inplace: bool
    ) -> Union[pd.DataFrame, None]:
        """
        Prepare and return resulted data frame with transformed target column.
        """
        if self.params["transformed_name"] is None:
            name = self.params["target_column"] + "_transformed"
        else:
            name = self.params["transformed_name"]
        df: pd.DataFrame = dataframe if inplace else dataframe.copy()
        df.loc[:, name] = new_target
        return df

    def _verbose(self, old_variance: float, new_variance: float) -> None:
        """
        Verbose method for transform operation Log.
        """
        part_of_variance: float = new_variance / (old_variance + AbstractVarianceReducer.EPSILON)
        log.info_log(f"After transformation {self}, the variance is {(part_of_variance * 100):.4f} % of the original")
        log.info_log(f"Variance transformation {old_variance:.4f} ===> {new_variance:.4f}")


def choose_on_table(alternatives: List[Any], dataframe) -> Any:
    """
    alternatives: [alternative_pandas, alternative_spark, ...]
    """
    if isinstance(dataframe, pd.DataFrame):
        return alternatives[0]
    elif isinstance(dataframe, types.SparkDataFrame):
        return alternatives[1]
    raise TypeError(f'Type of table must be one of {", ".join(AVAILABLE)}')


class DataframeHandler:
    @staticmethod
    def _handle_cases(__func_pandas: Callable, __func_spark: Callable, *args, **kwargs):
        """
        Helps handle cases with different types of dataframe in kwargs,
        available types - pandas, spark.
        """
        __func = choose_on_table([__func_pandas, __func_spark], kwargs[DATA])
        return __func(*args, **kwargs)

    @staticmethod
    def _handle_on_table(
        __func_pandas: Callable, __func_spark: Callable, variable: types.SparkOrPandas, *args, **kwargs
    ):
        """
        Helps handle cases with different types of dataframe as additional variable,
        available types - pandas, spark.
        """
        __func = choose_on_table([__func_pandas, __func_spark], variable)
        return __func(*args, **kwargs)


class SimpleDesigner(ABC, DataframeHandler):
    """
    Simple designer is the interface for designers for each dataframe and method.

    kwargs must contain parameter dataframe.
    """

    @abstractmethod
    def size_design(self, **kwargs) -> pd.DataFrame:
        pass

    @abstractmethod
    def effect_design(self, **kwargs) -> pd.DataFrame:
        pass

    @abstractmethod
    def power_design(self, **kwargs) -> pd.DataFrame:
        pass


class EmptyStratValue(Enum):
    NO_STRATIFICATION = 0


class StratificationUtil(ABC):
    """
    Stratification element
    https://en.wikipedia.org/wiki/Stratified_sampling
    """

    def __init__(self):
        self.strats = None

    @abstractmethod
    def fit(self, dataframe: types.SparkOrPandas, columns) -> None:
        pass

    @abstractmethod
    def strat_sizes(self) -> Dict[Any, int]:
        pass

    def is_trained(self) -> bool:
        """
        Returns whether the fit method was called.
        """
        return self.strats is not None

    def empty_strat(self) -> bool:
        """
        Check if there was no stratification.
        """
        return list(self.strats.keys()) == [EmptyStratValue.NO_STRATIFICATION]

    def _check_fit(self) -> None:
        """
        If fit method was not called before throw RuntimeError.
        """
        if not self.is_trained():
            raise RuntimeError("Call fit method !")

    def groups(self):
        """
        Returns
        -------
        items : Dict[tuple, SparkOrPandas] items
            iterable object (stratification value or group number, table for current stratification group)
        """
        self._check_fit()
        return self.strats.items()

    def size(self) -> int:
        """
        Calculate total rows amount considering filtering by threshold.

        Returns
        -------
        size: int
            Total size of filtered data
        """
        total_size: int = 0
        for size in self.strat_sizes().values():
            total_size += size
        return total_size

    def get_group_sizes(self, group_size: int) -> Dict[Any, int]:
        """
        Calculate size of group for each strat by total size of group.
        size(G_j) = group_size
        group_j_size = round(group_size * strat_size / total_size)

        Parameters
        ----------
        group_size : int
            Size for groups A, B, C ...

        Returns
        -------
        group_sizes : Dict[Any, int]
            Sizes for groups corresponding to stratification groups
        """
        filtered_size: int = self.size()
        strat_sizes: Dict[Tuple, int] = self.strat_sizes()
        group_sizes: Dict[Tuple, int] = {}
        for strat_value in self.strats:
            group_sizes[strat_value] = int(np.floor(group_size * strat_sizes[strat_value] / filtered_size))
        return group_sizes


class StatCriterion(ABC):
    """
    StatCriterion is the interface for arbitrary statistical criteria.
    """

    @abstractmethod
    def calculate_pvalue(self, group_a: Iterable[float], group_b: Iterable[float], **kwargs) -> np.ndarray:
        pass


class ABStatCriterion(StatCriterion):
    """
    ABStatCriterion is is the abstract class for statistical criterion used for design and test.
    """

    required_attributes = ["alias", "implemented_effect_types"]

    @classmethod
    def _send_type_error_msg(cls):
        error_msg = f"Choose effect_type from {cls.implemented_effect_types}"  # pylint: disable=E1101
        return error_msg

    @abstractmethod
    def calculate_effect(self, group_a: Iterable[float], group_b: Iterable[float], effect_type: str) -> np.ndarray:
        pass

    @abstractmethod
    def calculate_conf_interval(
        self, group_a: Iterable[float], group_b: Iterable[float], alpha: types.StatErrorType, effect_type: str, **kwargs
    ) -> List[Tuple]:
        pass

    def _make_ci(self, left_ci: np.ndarray, right_ci: np.ndarray, alternative: str) -> List:
        left_ci, right_ci = pvalue_pkg.choose_from_bounds(left_ci, right_ci, alternative)
        conf_intervals = list(zip(left_ci, right_ci))
        return conf_intervals

    def get_results(
        self,
        group_a: np.ndarray,
        group_b: np.ndarray,
        alpha: types.StatErrorType = 0.05,
        effect_type: str = "absolute",
        **kwargs,
    ) -> types.StatCriterionResult:
        return {
            "first_type_error": alpha,
            "pvalue": self.calculate_pvalue(group_a, group_b, effect_type=effect_type, **kwargs),
            "effect": self.calculate_effect(group_a, group_b, effect_type=effect_type),
            "confidence_interval": self.calculate_conf_interval(
                group_a, group_b, alpha=alpha, effect_type=effect_type, **kwargs
            ),
        }



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import functools
from typing import Iterable, List

import pandas as pd

from ambrosia import types

split_methods_list: List[str] = ["hash", "metric", "simple", "dim_decrease"]
metric_methods_list: List[str] = ["equal_size", "fast", "cluster"]
norm_list = ["l2", "l1"]


def check_type_decorator(type_checker=lambda set_value: set_value):
    """
    Decorator for setter method.
    If set_value is None => self.field = None
    Else using type_checker function set result of type_checker(set_value)
    """

    def inner_decorator(method):
        @functools.wraps(method)
        def wrapper(self, set_value):
            if set_value is None:
                method(self, None)
            else:
                method(self, type_checker(set_value))

        return wrapper

    return inner_decorator


def none_check_decorator(function):
    """
    Decorator for type checkers.
    If argument is None return None
    else return result of checker
    """

    @functools.wraps(function)
    def wrapper(argument):
        if argument is None:
            return None
        else:
            return function(argument)

    return wrapper


@none_check_decorator
def check_type_dataframe(dataframe: types.PassedDataType) -> types.PassedDataType:
    if isinstance(dataframe, str):
        if dataframe.endswith(".csv"):
            return pd.read_csv(dataframe)
        else:
            raise ValueError("``dataframe`` string value must be a link to .csv file")
    elif isinstance(dataframe, pd.DataFrame):
        return dataframe
    elif isinstance(dataframe, types.SparkDataFrame):
        return dataframe
    else:
        raise TypeError("``dataframe`` variable must be a pd.DataFrame object or a link to .csv file")


@none_check_decorator
def check_type_id_column(id_column: types.ColumnNameType) -> types.ColumnNameType:
    if isinstance(id_column, str):
        return id_column
    else:
        raise TypeError("id_column variable must be string")


@none_check_decorator
def check_type_id_columns(id_columns: types.ColumnNamesType) -> types.ColumnNamesType:
    if isinstance(id_columns, str):
        return [id_columns]
    elif isinstance(id_columns, Iterable):
        return id_columns
    else:
        raise TypeError("id_columns variable must be a string or an Iterable object")


@none_check_decorator
def check_type_group_size(groups_size: int) -> int:
    if isinstance(groups_size, int):
        return groups_size
    elif isinstance(groups_size, float):
        return int(groups_size)
    else:
        raise TypeError("groups_size variable must be int or float")


@none_check_decorator
def check_type_test_group_ids(test_group_ids: types.IndicesType) -> types.IndicesType:
    if isinstance(test_group_ids, Iterable):
        return test_group_ids
    else:
        raise TypeError("test_group_ids variable must be Iterable")


@none_check_decorator
def check_type_fit_columns(fit_columns: types.ColumnNamesType) -> types.ColumnNamesType:
    if isinstance(fit_columns, str):
        return [fit_columns]
    elif isinstance(fit_columns, Iterable):
        return fit_columns
    else:
        raise TypeError("fit_columns variable must be a string or an Iterable object")


@none_check_decorator
def check_type_strat_columns(strat_columns: types.ColumnNamesType) -> types.ColumnNamesType:
    if isinstance(strat_columns, str):
        return [strat_columns]
    elif isinstance(strat_columns, Iterable):
        return strat_columns
    else:
        raise TypeError("strat_columns variable must be a string or an Iterable object")


@none_check_decorator
def check_type_salt(salt: str) -> str:
    if isinstance(salt, str):
        return salt
    else:
        raise TypeError("salt variable must be a string")


@none_check_decorator
def check_split_method_value(split_method: str) -> str:
    if isinstance(split_method, str):
        if split_method in split_methods_list:
            return split_method
        else:
            raise ValueError(f'Choose correct split method, from {", ".join(split_methods_list)}')
    else:
        raise TypeError(f'method variable must be a string and from{", ".join(split_methods_list)}')


@none_check_decorator
def check_metric_method_value(method_metric: str) -> None:
    if isinstance(method_metric, str):
        if method_metric in metric_methods_list:
            return method_metric
        else:
            raise ValueError(f'Choose correct method_metric, from {", ".join(metric_methods_list)}')
    else:
        raise TypeError(f'method_metric variable must be a string and from{", ".join(metric_methods_list)}')


@none_check_decorator
def check_norm_value(norm: str) -> None:
    if isinstance(norm, str):
        if norm in norm_list:
            return norm
        else:
            raise ValueError(f'Choose correct norm, from {", ".join(norm_list)}')
    else:
        raise TypeError(f'norm variable must be a string and from{", ".join(norm_list)}')



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Iterable, List, Optional

import ambrosia.spark_tools.stratification as strat_pkg
from ambrosia import types
from ambrosia.tools import split_tools
from ambrosia.tools.import_tools import spark_installed

if spark_installed():
    import pyspark.sql.functions as spark_funcs
    from pyspark.sql import Window

HASH_COLUMN_NAME: str = "__hashed_ambrosia_column"
GROUPS_COLUMN: str = "group"
ROW_NUMBER: str = "__row_number"
EMPTY_VALUE: int = 0


def unite_spark_tables(*dataframes: types.SparkDataFrame) -> types.SparkDataFrame:
    """
    Union all spark dataframes.
    """
    amount_of_dataframes = len(dataframes)
    if not amount_of_dataframes:
        return None
    result = dataframes[0]
    for j in range(1, amount_of_dataframes):
        result = result.union(dataframes[j])
    return result


def add_hash_column(
    dataframe: types.SparkDataFrame,
    id_column: types.ColumnNameType,
    hash_function: str = "sha256",
    salt: Optional[str] = None,
) -> types.SparkDataFrame:
    """
    Returns new dataframe with column hashing id_column.

    Parameters
    ----------
    hash_function: str, default ``sha_256``
        Name of hash function
    """
    salt = "" if salt is None else salt
    column_with_id: types.SparkColumn = spark_funcs.concat(
        spark_funcs.col(id_column).cast("string"), spark_funcs.lit(salt)
    )
    if hash_function == "sha256":
        return dataframe.withColumn(HASH_COLUMN_NAME, spark_funcs.sha2(column_with_id, 256))
    elif hash_function == "sha512":
        return dataframe.withColumn(HASH_COLUMN_NAME, spark_funcs.sha2(column_with_id, 512))
    elif hash_function == "sha1":
        return dataframe.withColumn(HASH_COLUMN_NAME, spark_funcs.sha1(column_with_id))
    else:
        raise ValueError("Incorrect hash function name")


def get_hash_split(
    dataframe: types.SparkDataFrame,
    id_column: types.ColumnNameType,
    groups_size: int,
    labels: Iterable[str],
    groups_number: int = 2,
    # later group_b_indices: types.SparkDataFrame = None,
    hash_function: str = "sha256",
    salt: Optional[str] = None,
) -> types.SparkDataFrame:
    """
    Hash split.
    """
    hashed_dataframe = add_hash_column(dataframe, id_column, hash_function, salt)
    hashed_dataframe = hashed_dataframe.orderBy(HASH_COLUMN_NAME).limit(groups_number * groups_size)

    def udf_make_labels(row_number: int) -> str:
        label_ind = (row_number - 1) // groups_size
        return labels[label_ind]

    window = Window.orderBy(HASH_COLUMN_NAME).partitionBy(spark_funcs.lit(EMPTY_VALUE))
    result = hashed_dataframe.withColumn(ROW_NUMBER, spark_funcs.row_number().over(window)).withColumn(
        GROUPS_COLUMN, spark_funcs.udf(udf_make_labels)(spark_funcs.col(ROW_NUMBER))
    )
    result = result.drop(ROW_NUMBER, HASH_COLUMN_NAME)
    return result


def add_to_required_size(
    dataframe: types.SparkDataFrame,
    used_dataframe: types.SparkDataFrame,
    id_column: types.ColumnNameType,
    groups_size: int,
    current_sizes: List[int],
    labels: Iterable[str],
) -> types.SparkDataFrame:
    """
    Add elements for groups to required size.
    """
    not_used_ids: types.SparkDataFrame = dataframe.join(used_dataframe, on=id_column, how="leftanti")
    required_sizes: List[int] = [groups_size - size_ for size_ in current_sizes]
    total_required: int = sum(required_sizes)
    not_used_ids = not_used_ids.limit(total_required)

    # Now it's linear search, probably there will be not so many groups
    def udf_make_labels_with_find(row_number: int):
        current_total: int = 0
        for j in range(len(required_sizes)):
            current_total += required_sizes[j]
            if not required_sizes[j]:
                continue
            if row_number <= current_total:
                return labels[j]

    return (
        not_used_ids.withColumn(
            ROW_NUMBER,
            spark_funcs.row_number().over(
                Window.orderBy(spark_funcs.lit(EMPTY_VALUE)).partitionBy(spark_funcs.lit(EMPTY_VALUE))
            ),
        )
        .withColumn(GROUPS_COLUMN, spark_funcs.udf(udf_make_labels_with_find)(ROW_NUMBER))
        .drop(ROW_NUMBER)
    )


def get_split(
    dataframe: types.SparkDataFrame,
    split_method: str,
    id_column: types.ColumnNameType,
    groups_size: int,
    groups_number: int = 2,
    # later group_b_indices: types.SparkDataFrame = None,
    strat_columns: Optional[List] = None,
    hash_function: str = "sha256",
    salt: Optional[str] = None,
    labels: Optional[Iterable[str]] = None,
) -> types.SparkDataFrame:
    """
    Get split.
    """
    total_size: int = dataframe.count()

    if groups_number * groups_size > total_size:
        raise ValueError("Total sample size is more, than shape of table")

    if total_size > dataframe.dropDuplicates([id_column]).count():
        raise ValueError(f"Id column {id_column} contains duplicates, ids must be unique for split")

    if labels is None:
        labels = split_tools.make_labels_for_groups(groups_number)

    stratification = strat_pkg.Stratification()
    stratification.fit(dataframe, strat_columns)
    strat_sizes = stratification.get_group_sizes(groups_size)
    tables_on_stratification = []

    # List of sizes for each group, better to make without calling to cluster
    current_sizes: List[int] = [0] * groups_number

    for strat_value, strat_table in stratification.groups():
        current_size: int = strat_sizes[strat_value]
        if split_method == "hash":
            current_table = get_hash_split(
                strat_table,
                id_column,
                current_size,
                labels,
                groups_number,
                hash_function,
                salt,
            )
        else:
            raise ValueError("Split method is not found")
        tables_on_stratification.append(current_table)
        for j in range(groups_number):
            current_sizes[j] += current_size

    # Unite dataframes from each stratification group
    used_ids: types.SparkDataFrame = unite_spark_tables(*tables_on_stratification)
    additional_table: types.SparkDataFrame = add_to_required_size(
        dataframe, used_ids, id_column, groups_size, current_sizes, labels
    )
    return unite_spark_tables(used_ids, additional_table)



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional

import numpy as np
import scipy.stats as sps

from ambrosia import types


def __helper_calc_empirical_power(conf_interval: types.ManyIntervalType) -> np.ndarray:
    """
    Calculate esimation of power / first type error.

    Parameters
    ----------
    conf_interval : Tuple[np.ndarray]
        conf_interval[0] - left boundaries, conf_interval[1] - right
        shapes - (amount_of_samples, size_of_parameter_grid)

    Returns
    -------
    power : np.ndarray
        power / first type error for each parameter in parameter grid
    """
    left, right = conf_interval
    mask: np.ndarray = (left <= 0) & (0 <= right)
    power: np.ndarray = 1 - mask.mean(axis=0)
    return power


def __helper_bin_search_for_size(
    interval_type: str, confidence_level: float, p_a: float, p_b: float, amount: int, power: float
) -> int:
    """
    Make binary search for size to gain given power.

    Parameters
    ----------
    interval_type : str
        Type of confidence interval
    confidence_level : float
        1 - first type error value
    p_a : Iterable[float]
        Conversion in A group
    p_b : Iterable[float]
        Conversion in B group
    amount : int
        Amount of generated samples for one n(trials amount), to estimate power
    power : float
        Desired level of power

    Returns
    -------
    sample_size : int
        Such sample size, that for arguments confidence interval give
        satisfying power
    """

    def power_helper(trials: int) -> float:
        import ambrosia.tools.bin_intervals as bi

        sample_a = sps.binom.rvs(n=trials, p=p_a, size=amount)
        sample_b = sps.binom.rvs(n=trials, p=p_b, size=amount)
        binom_kwargs = {
            "interval_type": interval_type,
            "a_success": sample_a,
            "b_success": sample_b,
            "a_trials": trials,
            "b_trials": trials,
            "confidence_level": confidence_level,
        }
        conf_interval: types.IntervalType = bi.BinomTwoSampleCI.confidence_interval(**binom_kwargs)
        return __helper_calc_empirical_power(conf_interval)

    # Find upper bound for size
    k: int = 10
    current_power: float = 0
    while current_power < power:
        current_power = power_helper(trials=2**k)
        k += 1
    right: int = 2**k
    left: int = 1

    while right - left > 0:
        middle: int = (left + right) // 2
        current_power = power_helper(middle)
        if current_power < power:
            left = middle
        else:
            right = middle
    return right


def __helper_bin_search_for_delta(
    interval_type: str,
    confidence_level: float,
    p_a: float,
    trials: int,
    amount: int,
    power: float,
    epsilon: float = 0.0001,
) -> Optional[float]:
    """
    Make binary search for delta to gain given power for
    get_table_effect_on_sample_size function

    Parameters
    ----------
    interval_type : str
        Type of confidence interval
    confidence_level : float
        1 - first type error value
    p_a : Iterable[float]
        Conversion in A group
    trials: int
        Number of trials in groups
    amount : int
        Amount of generated samples for one n(trials amount), to estimate power
    power : float
        Desired level of power
    epsilon: float, default : ``0.001``
        Precision for binary search solution

    Returns
    -------
    delta : Optional[float]
        |delta - delta_optimal| < epsilon
        None if there are no satisfying deltas
    """

    def power_helper(delta: float) -> float:
        import ambrosia.tools.bin_intervals as bi

        sample_a = sps.binom.rvs(n=trials, p=p_a, size=amount)
        p_b: float = p_a - delta
        sample_b = sps.binom.rvs(n=trials, p=p_b, size=amount)
        binom_kwargs = {
            "interval_type": interval_type,
            "a_success": sample_a,
            "b_success": sample_b,
            "a_trials": trials,
            "b_trials": trials,
            "confidence_level": confidence_level,
        }
        conf_interval: types.IntervalType = bi.BinomTwoSampleCI.confidence_interval(**binom_kwargs)
        return __helper_calc_empirical_power(conf_interval)

    current_delta: float = epsilon
    current_power: float = 0
    mult_coef: float = -1 if p_a <= 0.5 else 1
    while current_power < power:
        current_power = power_helper(delta=mult_coef * current_delta)
        current_delta *= 2
        p_b: float = p_a - mult_coef * current_delta
        if (p_b < 0 or p_b > 1) and (current_delta != p_a):
            current_delta = p_a
            continue
        if p_b < 0 or p_b > 1:
            return None

    right: float = current_delta
    left: float = 0

    while right - left > epsilon:
        middle: float = (left + right) / 2
        current_power = power_helper(mult_coef * middle)
        if current_power < power:
            left = middle
        else:
            right = middle
    return mult_coef * right



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Groups splitting methods.

Module contains `Splitter` core class and `split` method which are
intended to solve group splitting problems, primarily for A/B/.. tests.
Group splitting tasks usually include following parameters: number of groups,
group sizes, and splitting algorithm.

Currently, group splitting problems could be solved using data provided
in form of both pandas and Spark(with some restrictions) dataframes.

"""
from __future__ import annotations

from typing import Optional

import yaml

from ambrosia import types
from ambrosia.tools import log, type_checks
from ambrosia.tools.ab_abstract_component import ABMetaClass, ABToolAbstract

from .handlers import data_shape, handle_full_split, split_data_handler

SPLITTING_BOUND_CONST: float = 0.5


class Splitter(yaml.YAMLObject, ABToolAbstract, metaclass=ABMetaClass):
    """
    Unit for creating experimental groups from batch data.

    Split your data into groups of selected size with respect to:
        - Stratification columns
        - Metric distance of objects in feature space
        - Set of passed ids

    Parameters
    ----------
    dataframe : PassedDataType, optional
        Dataframe or string name of .csv table which
        contains data used for groups split.
    id_column : IdColumnNameType, optional
        Name of id column which is used in hash split.
    groups_size : int, optional
        Size of the splitted groups.
    test_group_ids : PeriodColumnNamesType, optional
        Ids of objects which are in B(test) group.
        Used in tasks of post experiment A(control) group pick up.
    fit_columns : PeriodColumnNamesType, optional
        List of columns names which values will be interpreted as
        coordinates of points in multidimensional space during metric split.
    strat_columns : PeriodColumnNamesType, optional
        Columns for stratification.
        https://en.wikipedia.org/wiki/Stratified_sampling

    Attributes
    ----------
    dataframe : PassedDataType
        Pandas or Spark dataframe with split data.
    id_column : IdColumnNameType
        Name of id column which is used in hash split.
    groups_size : int
        Split size of groups.
    test_group_ids : PeriodColumnNamesType
        Ids of objects which are in B(test) group.
    fit_columns : PeriodColumnNamesType
        List of columns names used for metric split.
    strat_columns : PeriodColumnNamesType
        Stratification columns names.

    Examples
    --------
    Our development team decided to add onboarding to the mobile app.
    Already knowing the required group size, we would like to select
    users for groups A and B respectively. Using the splitter class,
    this task could be done in the following way:

    >>> splitter = Splitter(dataframe=dataframe)
    >>> splitter.run(group_size=1000, method='hash', salt='onboarding')

    Suppose now, we know that people of different ages and from several
    countries use our application, so we would like to take this into
    account during split. To do this, you might use stratification,
    which can be easily applied by passing only one additional parameter:

    >>> splitter = Splitter(data=dataframe, strat_columns=['age', 'country'])
    >>> splitter.run(group_size=1000, method='hash', salt='onboarding')

    If we have fixed users for the testing group,
    this can be specified as a parameter:

    >>> splitter = Splitter(data=dataframe, strat_columns=['age', 'country'])
    >>> splitter.run(method='hash',
    >>>              salt='onboarding',
    >>>              test_group_ids=B_group_id
    >>> )

    Notes
    -----

    Main methods for split:

    Simple:
        - Randomly chosen groups (via ``np.random.choice``).

    Hash:
        - Using hashing of identifiers and distribution
          by buckets, selects the desired buckets for groups formation.

    Metric:
        - For a fixed reference group or a randomly selected one,
          other groups are selected using the nearest neighbor method
          (for desired list of columns passed in ``fit_columns`` parameter).

    Constructors:

        >>> # Empty constructor
        >>> splitter = Splitter()
        >>> # Some data
        >>> splitter = Splitter(dataframe=df,
        >>>                     id_column='my_id_column',
        >>>                     strat_columns=['gender', 'age'],
        >>>                     test_group_ids=ids_for_B_group
        >>> )

    Setters:

        >>> splitter.set_dataframe(dataframe)
        >>> # You can pass string for pd.read_csv
        >>> splitter.set_dataframe('name_of_table.csv')
        >>> # Other setters
        >>> splitter.set_group_size(1000)
        >>> splitter.set_strat_columns(['age', 'region'])

    Run:

        >>> splitter.run(method='hash', groups_size=10000)
        >>> splitter.run(method='metric'
        >>>              test_group_ids=b_group,
        >>>              id_column='id',
        >>>              strat_columns=['age', 'city']
        >>>              fit_columns=['metric_history_column', 'other_metric']
        >>>              method_meric='fast', # It is used as kwarg
        >>>              norm='l2' # It is used as kwarg
        >>> )

    Load from yaml config:

    >>> config = '''
                !splitter # <--- this is yaml tag (important!)
                    groups_size:
                        1000
                    id_column:
                        id
                    strat_columns:
                        - age
                        - country
            '''
    >>> splitter = yaml.load(config)
    >>> # Or use the implmented function
    >>> splitter = load_from_config(config)
    """

    yaml_tag = "!splitter"

    @type_checks.check_type_decorator(type_checks.check_type_dataframe)
    def set_dataframe(self, dataframe: Optional[types.PassedDataType]) -> None:
        self.__df = dataframe

    @type_checks.check_type_decorator(type_checks.check_type_id_column)
    def set_id_column(self, id_column: Optional[str]) -> None:
        self.__id_column = id_column

    @type_checks.check_type_decorator(type_checks.check_type_group_size)
    def set_group_size(self, groups_size: Optional[int]) -> None:
        self.__groups_size = groups_size

    @type_checks.check_type_decorator(type_checks.check_type_test_group_ids)
    def set_test_group_ids(self, test_group_ids: types.IndicesType) -> None:
        self.__test_group_ids = test_group_ids

    @type_checks.check_type_decorator(type_checks.check_type_fit_columns)
    def set_fit_columns(self, fit_columns: types.ColumnNamesType) -> None:
        self.__fit_columns = fit_columns

    @type_checks.check_type_decorator(type_checks.check_type_strat_columns)
    def set_strat_columns(self, strat_columns: types.ColumnNamesType) -> None:
        self.__strat_columns = strat_columns

    def __init__(
        self,
        dataframe: Optional[types.PassedDataType] = None,
        id_column: Optional[types.ColumnNameType] = None,
        groups_size: Optional[int] = None,
        test_group_ids: Optional[types.IndicesType] = None,
        fit_columns: Optional[types.ColumnNamesType] = None,
        strat_columns: Optional[types.ColumnNamesType] = None,
    ):
        """
        Splitter class constructor to initialize the object.
        """
        self.set_dataframe(dataframe)
        self.set_id_column(id_column)
        self.set_group_size(groups_size)
        self.set_test_group_ids(test_group_ids)
        self.set_fit_columns(fit_columns)
        self.set_strat_columns(strat_columns)

    def __getstate__(self):
        """
        Get the state of the object to serialize.
        """
        return dict(
            id_column=self.__id_column,
            groups_size=self.__groups_size,
            fit_columns=self.__fit_columns,
            strat_columns=self.__strat_columns,
        )

    @classmethod
    def from_yaml(cls, loader: yaml.Loader, node: yaml.Node):
        kwargs = loader.construct_mapping(node)
        return cls(**kwargs)

    def run(
        self,
        method: str,
        dataframe: Optional[types.PassedDataType] = None,
        id_column: Optional[types.ColumnNameType] = None,
        groups_size: Optional[int] = None,
        part_of_table: Optional[float] = None,
        groups_number: int = 2,
        test_group_ids: Optional[types.IndicesType] = None,
        strat_columns: Optional[types.ColumnNamesType] = None,
        salt: Optional[str] = None,
        fit_columns: Optional[types.ColumnNamesType] = None,
        **kwargs,
    ) -> types.SplitterResult:
        """
        Perform a split into groups with selected or saved parameters.

        Parameters
        ----------
        method : str
            Split method, for example ``"hash"``.
        dataframe : PassedDataType, optional
            Dataframe or string name of .csv table which
            contains data used for groups split.
        id_column : IdColumnNameType, optional
             Name of id column which is used in hash split.
        groups_size : int, optional
            Size of the splitted groups.
        part_of_table: float, optional
            Split factor(for group A) for tasks of dataframe full split.
            If is not ``None``, then overrides ``groups_size`` parameter
            during the split.
        groups_number : int, default: ``2``
            Number of groups to be splitted.
        test_group_ids : PeriodColumnNamesType, optional
            Ids of objects which are in B(test) group.
            Used in tasks of post experiment A(control) group pick up.
        strat_columns : PeriodColumnNamesType, optional
            Columns for stratification.
            https://en.wikipedia.org/wiki/Stratified_sampling
        salt : str, optional
            Salt for hashing in hash-split.
        fit_columns : PeriodColumnNamesType, optional
            List of columns names which values will be interpreted as
            coordinates of points in multidimensional space during metric split.
        **kwargs : Dict
            Other keyword arguments.

        Returns
        -------
        groups : pd.DataFrame
            Returns a dataframe with groups and label column.
            Dataframe will contain all columns of the original dataframe.

        Other Parameters
        ----------------
        threads : int, default : ``1``
            Number of threads used for calculations.
        """
        method: str = type_checks.check_split_method_value(method)
        dataframe: types.PassedDataType = type_checks.check_type_dataframe(dataframe)
        id_column: types.ColumnNameType = type_checks.check_type_id_column(id_column)
        groups_size: int = type_checks.check_type_group_size(groups_size)
        test_group_ids: types.IndicesType = type_checks.check_type_test_group_ids(test_group_ids)
        fit_columns: types.ColumnNamesType = type_checks.check_type_fit_columns(fit_columns)
        strat_columns: types.ColumnNamesType = type_checks.check_type_strat_columns(strat_columns)

        arguments_choice: types._PrepareArgumentsType = {
            "dataframe": (self.__df, dataframe),
        }

        strat_columns: str = strat_columns if strat_columns is not None else self.__strat_columns
        test_group_ids = test_group_ids if test_group_ids is not None else self.__test_group_ids
        id_column = id_column if id_column is not None else self.__id_column

        if test_group_ids is not None:
            arguments_choice["group_b_indices"] = (None, test_group_ids)
        else:
            arguments_choice["groups_size"] = (self.__groups_size, groups_size)
        if part_of_table is not None:
            # Group size will be set later
            arguments_choice["groups_size"] = (self.__groups_size, 0)
            if groups_size is not None:
                log.info_log("Groups size variable ignored because part splitting variable set")
            if groups_number > 2:
                groups_number = 2
                log.info_log("Groups number was set to 2 because part splitting variable set")

        if method in ("metric", "dim_decrease"):
            # For methods use metric/cluster/unsupervised approach
            arguments_choice["fit_columns"] = (self.__fit_columns, fit_columns)

        chosen_args: types._UsageArgumentsType = Splitter._prepare_arguments(arguments_choice)

        if part_of_table is not None:
            split_part: float = part_of_table if (part_of_table <= SPLITTING_BOUND_CONST) else 1 - part_of_table
            chosen_args["groups_size"] = round(split_part * data_shape(chosen_args["dataframe"]))

        chosen_args["split_method"] = method
        chosen_args["id_column"] = id_column
        chosen_args["strat_columns"] = strat_columns
        chosen_args["salt"] = salt
        chosen_args["groups_number"] = groups_number
        groups: types.SplitterResult = split_data_handler(**chosen_args, **kwargs)

        if part_of_table is not None:
            return handle_full_split(chosen_args["dataframe"], groups, part_of_table, id_column)

        return groups


def load_from_config(yaml_config: str, loader: type = yaml.Loader) -> Splitter:
    """
    Restore a ``Splitter`` class instance from a yaml config.

    For yaml_config parameter you can pass file name with
    config, which must ends with .yaml, for example: "config.yaml".
    For loader you can choose SafeLoader.
    """
    if isinstance(yaml_config, str) and yaml_config.endswith(".yaml"):
        with open(yaml_config, "r", encoding="utf-8") as file:
            return yaml.load(file, Loader=loader)
    return yaml.load(yaml_config, Loader=loader)


def split(
    method: str,
    dataframe: Optional[types.PassedDataType] = None,
    id_column: Optional[types.ColumnNameType] = None,
    groups_size: Optional[int] = None,
    part_of_table: Optional[float] = None,
    groups_number: int = 2,
    test_group_ids: Optional[types.IndicesType] = None,
    strat_columns: Optional[types.ColumnNamesType] = None,
    salt: Optional[str] = None,
    fit_columns: Optional[types.ColumnNamesType] = None,
    threads: int = 1,
    **kwargs,
) -> types.SplitterResult:
    """
    Function wrapper around the ``Splitter`` class.

    Used to create splitted groups from the dataframe.

    Creates an instance of the ``Splitter`` class internally and execute
    run method with corresponding arguments.

    Parameters
    ----------
    method : str
        Split method, for example ``"hash"``.
    dataframe : PassedDataType, optional
        Dataframe or string name of .csv table which
        contains data used for groups split.
    id_column : IdColumnNameType, optional
            Name of id column which is used in hash split.
    groups_size : int, optional
        Size of the splitted groups.
    part_of_table: float, optional
        Split factor(for group A) for tasks of dataframe full split.
        If is not ``None``, then overrides ``groups_size`` parameter
        during the split.
    groups_number : int, default : ``2``
        Number of groups to be splitted.
    test_group_ids : PeriodColumnNamesType, optional
        Ids of objects which are in B(test) group.
        Used in tasks of post experiment A(control) group pick up.
    strat_columns : PeriodColumnNamesType, optional
        Columns for stratification.
        https://en.wikipedia.org/wiki/Stratified_sampling
    salt : str, optional
        Salt for hashing in hash-split.
    fit_columns : PeriodColumnNamesType, optional
        List of columns names which values will be interpreted as
        coordinates of points in multidimensional space during metric split.
    threads : int, default : ``1``
        Number of threads used for calculations.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    groups : pd.DataFrame
        Returns a dataframe with groups and label column.
        Dataframe will contain all columns of the original dataframe.
    """
    return Splitter(
        dataframe=dataframe,
        id_column=id_column,
        groups_size=groups_size,
        fit_columns=fit_columns,
        test_group_ids=test_group_ids,
        strat_columns=strat_columns,
    ).run(method, salt=salt, threads=threads, part_of_table=part_of_table, groups_number=groups_number, **kwargs)



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Subpackage for data preprocessing, including methods for accelerating
experiments.
"""
from .aggregate import AggregatePreprocessor
from .cuped import Cuped, MultiCuped
from .ml_var_reducer import MLVarianceReducer
from .preprocessor import Preprocessor
from .robust import IQRPreprocessor, RobustPreprocessor
from .transformers import BoxCoxTransformer, LogTransformer

__all__ = [
    "AggregatePreprocessor",
    "Cuped",
    "MultiCuped",
    "MLVarianceReducer",
    "Preprocessor",
    "RobustPreprocessor",
    "IQRPreprocessor",
    "BoxCoxTransformer",
    "LogTransformer",
]



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, Iterable, List

import numpy as np
import pandas as pd
from joblib import Parallel, delayed, parallel_backend
from tqdm.auto import tqdm

import ambrosia.spark_tools.theory as th_pkg
import ambrosia.tools._lib._bootstrap_tools as solver_pkg
import ambrosia.tools._lib._selection_aide as select_pkg
from ambrosia import types
from ambrosia.tools.import_tools import spark_installed

if spark_installed():
    import pyspark.sql.functions as spark_functions

BOOSTRAP_BASE_CONST: int = 10
RANDOM_SAMPLE_SEED: int = 42
FIRST_TYPE_ERROR: float = 0.05
THREADS_BOOTSTRAP: int = 2  # Creates a significant reduction in runtime
N_JOBS_MULTIPROCESS: int = 1
ACCEPTED_CRITERIA: List[str] = ["ttest"]
BOOTSTRAP_BACKEND: str = "threading"
ROUND_DIGITS_PERCENT: int = 1


def inject_effect(dataframe: types.SparkDataFrame, column: types.ColumnNameType, effect: float) -> types.SparkDataFrame:
    """
    Injects effect to column of given dataframe and returns injected one.

    Injection conducts via adding mean * delta_relative_effect.

    Parameters
    ----------
    dataframe : Spark dataframe
        Table with column, where efect will be injected
    column : Column type
        Column which will be used
    effect : float
        Value of effect, for example for 20% percent growth pass 1.2

    Returns
    -------
    effected_dataframe : Spark dataframe
        Table with changed column
    """
    multiplicator: float = effect - 1
    current_mean, _ = th_pkg.get_stats_from_table(dataframe, column)
    return dataframe.withColumn(column, spark_functions.col(column) + current_mean * multiplicator)


def evaluate_criterion(
    dataframe: types.SparkDataFrame,
    column: types.ColumnNameType,
    effect: float,
    group_size: int,
    alpha: float = FIRST_TYPE_ERROR,
    criterion: str = ACCEPTED_CRITERIA[0],
) -> int:
    """
    Evaluate criterion, returns 0 if H1 is not rejected, 1 otherwise.

    Check hypotesis for given column, if effect injected for given sample size.
    H0: means in two groups equals, H1: means are not equal

    Parameters
    ----------
    dataframe : Spark dataframe
        Table with column, which wiil be used
    column : Column type
        Column with metric, which will be used in criterion
    effect : float
        Value of effect to be tested, for example, for 20% effect pass 1.2
    group_size : int
        Size for each of two groups, which wiil be sampled
    alpha : float
        Bound for first type error, will be used as pvalue <= alpha
    criterion : str
        Name of criterion, default ttest, see list acceptable criteria

    Returns
    -------
    is_rejected : bool
        Is H1 correct
    """
    total_size: int = dataframe.count()
    if group_size * 2 > total_size:
        err_msg: str = "Total sampled values more than table size"
        raise ValueError(err_msg)

    part: float = 2 * group_size / total_size
    data_a, data_b = dataframe.sample(part).randomSplit([0.5, 0.5], seed=RANDOM_SAMPLE_SEED)
    data_b = inject_effect(data_b, column, effect)

    if criterion == "ttest":
        _, pvalue = th_pkg.ttest_spark(data_a, data_b, column)
    else:
        err_msg: str = f"Choose criterion from {ACCEPTED_CRITERIA}"
        raise ValueError(err_msg)
    return pvalue < alpha


def calc_empiric_power(
    dataframe: types.SparkDataFrame,
    column: types.ColumnNameType,
    effect: float,
    group_size: int,
    first_error: float = FIRST_TYPE_ERROR,
    bootstrap_size: int = BOOSTRAP_BASE_CONST,
    criterion: str = ACCEPTED_CRITERIA[0],
    threads: int = THREADS_BOOTSTRAP,
) -> float:
    """
    Calculate empiric power of criterion via thread pool.

    Parameters
    ----------
    dataframe : Spark dataframe
        Table with column, which wiil be used
    column : Column type
        Column with metric, which will be used in criterion
    effect : float
        Value of effect to be tested, for example, for 20% effect pass 1.2
    group_size: int
        Size for each of two groups, which wiil be sampled
    first_error : float
        Bound for first type error, will be used as pvalue <= first_error
    bootstrap_size : int
        Amount of groups to be sampled
    criterion : str
        Name of criterion, default ttest, see list acceptable criteria
    threads : int
        Amount of threads used in thread pool

    Returns
    -------
    empirical_power : float
        Empirical power, calculated as frequecy of rejected hypotesis
    """
    if threads > 1:
        with parallel_backend(BOOTSTRAP_BACKEND, n_jobs=threads):
            exp_results = Parallel(verbose=False)(
                delayed(evaluate_criterion)(
                    dataframe=dataframe,
                    column=column,
                    effect=effect,
                    group_size=group_size,
                    alpha=first_error,
                    criterion=criterion,
                )
                for _ in range(bootstrap_size)
            )
    else:
        exp_results = []
        for _ in range(bootstrap_size):
            exp_results.append(
                evaluate_criterion(
                    dataframe=dataframe,
                    column=column,
                    effect=effect,
                    alpha=first_error,
                    group_size=group_size,
                    criterion=criterion,
                )
            )
    return np.mean(exp_results)


def get_table_power(
    dataframe: types.SparkDataFrame,
    metrics: Iterable[types.ColumnNameType],
    effects: Iterable[float],
    group_sizes: Iterable[int],
    alphas: Iterable[float],
    bootstrap_size: int = BOOSTRAP_BASE_CONST,
    threads: int = THREADS_BOOTSTRAP,
    use_tqdm: bool = True,
    as_numeric: bool = False,
) -> types.DesignerResult:
    """
    Calculate table of criterion empirical power with rows effects and columns sizes.

    Parameters
    ----------
    dataframe : Spark dataframe
        Table with column, which wiil be used
    metrics : Iterable of column type
        Iterable set of columns for designing
    effects : Iterable[float]
        List of effects which we want to check
    group_sizes : Iterable[int]
        List of group sizes which we want to check
    alphas : Iterable[float]
        1st type error bound, passed as list, for example [0.05]
    bootstrap_size: int, default: ``10``
        Amount of pairs of groups A/B to be sampled for estimation power
    threads : int
        Amount of threads for thread pool
    use_tqdm : bool
        Whether to use progress bar
    as_numeric : bool, default False
        Whether to return a number or a string with percentages

    Returns
    -------
    report : Union[pd.DataFrame, Dict[str, pd.DataFrame]
        Tables with sizes for group A, group B, effects, erros and metrics names
        Effects for indices
        Group sizes for columns
        table for each metric: dict[metric name] = table with mde(effects)
        or one table, if only one metric passed
    """
    if len(alphas) > 1:
        raise ValueError("For power table you can pass only one first error bound")
    result = pd.DataFrame(columns=group_sizes, index=effects)
    iterate_params = tqdm(zip(effects, group_sizes)) if use_tqdm else zip(effects, group_sizes)
    for column in metrics:
        for effect, group_size in iterate_params:
            power = calc_empiric_power(
                dataframe=dataframe,
                column=column,
                effect=effect,
                group_size=group_size,
                first_error=alphas[0],
                bootstrap_size=bootstrap_size,
                threads=threads,
            )
            if as_numeric:
                result.loc[effect, group_size] = power
            else:
                result.loc[effect, group_size] = (round(power * 100, ROUND_DIGITS_PERCENT)).astype(str) + "%"
    result.columns.name = "sample sizes"
    result.index.name = "effect"
    return result


def optimize_group_size(
    dataframe: types.SparkDataFrame,
    column: types.ColumnNameType,
    effect: float,
    beta: float,
    first_error: float = FIRST_TYPE_ERROR,
    bootstrap_size: int = BOOSTRAP_BASE_CONST,
    threads: int = THREADS_BOOTSTRAP,
) -> int:
    """
    Optimize group size for fixed effect and errors using empiric solution.
    Spark requests are made using thread pool.

    Parameters
    ----------
    dataframe : Spark table
        Table for designing experiment
    column : Column type
        Column, containg metric for designing
    effect : float
        Size of group both groups
    beta: float
        2nd type error bound
    first_error : float, default: ``(0.05,)``
        1st type error bound
    bootstrap_size : int, default: ``10``
        Amount of pairs of groups A/B to be sampled for estimation power
    threads : int
        Amount of threads for thread pool

    Returns
    -------
    optimal_size : int
        Groups sizes calculted via empiric power optimization
    """
    power: float = 1 - beta
    solver = solver_pkg.EmpiricSizeSolution(calc_empiric_power, power, ["group_size"])
    return solver.calc_binary(
        dataframe=dataframe,
        column=column,
        effect=effect,
        first_error=first_error,
        bootstrap_size=bootstrap_size,
        threads=threads,
    )


def get_table_size(
    dataframe: types.SparkDataFrame,
    metrics: Iterable[types.ColumnNameType],
    effects: Iterable[float],
    betas: Iterable[float],
    alphas: Iterable[float],
    bootstrap_size: int = BOOSTRAP_BASE_CONST,
    threads: int = THREADS_BOOTSTRAP,
    n_jobs: int = N_JOBS_MULTIPROCESS,
    use_tqdm: bool = True,
    as_numeric: bool = False,
) -> types.DesignerResult:
    """
    Find sizes by variating other params for many columns using thread pool for power estimation.

    Parameters
    ----------
    dataframe : Spark table
        Table for designing experiment
    metrics : Iterable of column type
        Iterable set of columns for designing
    effects : Iterable[float]
        List of group size which we want to check
    betas : Iterable[float]
        2nd type error bounds
    alpha : Iterable[float]
        1st type error bounds
    bootstrap_size : int, default: ``10``
        Amount of pairs of groups A/B to be sampled for estimation power
    threads : int
        Amount of threads for thread pool
    n_jobs : int
        Amount of jobs for metrics variating
    use_tqdm : bool
        Whether to use progress bar
    as_numeric : bool, default False
        Whether to return a number or a string with percentages

    Returns
    -------
    report : Union[pd.DataFrame, Dict[str, pd.DataFrame]
        Tables with sizes for group A, group B, effects, erros and metrics names
        Effects for indices
        (alpha(1 type error), beta(2 type error)) for columns
        table for each metric: dict[metric name] = table with mde(effects),
        or one table if one metric passed
    """
    params: Dict[str, Iterable[Any]] = {"effect": effects, "beta": betas, "first_error": alphas}
    results: Dict[types.ColumnNameType, pd.DataFrame] = {}
    for column_name in metrics:
        selector = select_pkg.Selector(
            optimize_group_size,
            params,
            n_jobs,
            use_tqdm,
            dataframe=dataframe,
            column=column_name,
            bootstrap_size=bootstrap_size,
            threads=threads,
        )
        results[column_name] = selector.get_table_size(as_numeric)
    return results[metrics[0]] if len(results) == 1 else results


def optimize_effect(
    dataframe: types.SparkDataFrame,
    column: types.ColumnNameType,
    group_size: int,
    beta: float,
    first_error: float = FIRST_TYPE_ERROR,
    bootstrap_size: int = BOOSTRAP_BASE_CONST,
    threads: int = THREADS_BOOTSTRAP,
) -> int:
    """
    Optimize effect for fixed size and errors using empiric solution.
    Spark requests are made using thread pool.

    Parameters
    ----------
    dataframe : Spark table
        Table for designing experiment
    column : Column type
        Column, containg metric for designing
    group_size : int
        Size of group both groups
    beta: float
        2nd type error bound
    first_error : float, default: ``(0.05,)``
        1st type error bound
    bootstrap_size : int, default: ``10``
        Amount of pairs of groups A/B to be sampled for estimation power
    threads : int
        Amount of threads for thread pool
    """
    power: float = 1 - beta
    solver = solver_pkg.EmpiricEffectSolution(calc_empiric_power, power, "effect")
    return solver.calc_binary(
        dataframe=dataframe,
        column=column,
        group_size=group_size,
        first_error=first_error,
        bootstrap_size=bootstrap_size,
        threads=threads,
    )


def get_table_effect(
    dataframe: types.SparkDataFrame,
    metrics: Iterable[types.ColumnNameType],
    group_sizes: Iterable[int],
    betas: Iterable[float],
    alphas: Iterable[float],
    bootstrap_size: int = BOOSTRAP_BASE_CONST,
    threads: int = THREADS_BOOTSTRAP,
    n_jobs: int = N_JOBS_MULTIPROCESS,
    use_tqdm: bool = True,
    as_numeric: bool = False,
) -> types.DesignerResult:
    """
    Find effects by variating other params for many columns using thread pool for power estimation.

    Parameters
    ----------
    dataframe : Spark table
        Table for designing experiment
    metrics : Iterable of column type
        Iterable set of columns for designing
    group_sizes : Iterable[int]
        List of group size which we want to check
    betas : Iterable[float]
        2nd type error bounds
    alphas : Iterable[float]
        1st type error bounds
    bootstrap_size : int, default: ``10``
        Amount of pairs of groups A/B to be sampled for estimation power
    threads : int
        Amount of threads for thread pool
    n_jobs : int
        Amount of jobs for metrics variating
    use_tqdm : bool
        Whether to use progress bar
    as_numeric : bool, default False
        Whether to return a number or a string with percentages

    Returns
    -------
    report : Union[pd.DataFrame, Dict[str, pd.DataFrame]
        Tables with sizes for group A, group B, effects, erros and metrics names
        Group sizes for indices
        (alpha(1 type error), beta(2 type error)) for columns
        table for each metric: dict[metric name] = table with mde(effects),
        or one table if one metric passed
    """
    params: Dict[str, Iterable[Any]] = {"group_size": group_sizes, "beta": betas, "first_error": alphas}
    results: Dict[types.ColumnNameType, pd.DataFrame] = {}
    for column_name in metrics:
        selector = select_pkg.Selector(
            optimize_effect,
            params,
            n_jobs,
            use_tqdm,
            dataframe=dataframe,
            column=column_name,
            bootstrap_size=bootstrap_size,
            threads=threads,
        )
        results[column_name] = selector.get_table_effect(as_numeric)
    return results[metrics[0]] if len(results) == 1 else results



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Ambrosia
===============================

Ambrosia is a Python library for A/B tests design, split and effect
measurement. It provides rich set of methods for conducting full
A/B test pipeline. In particular, a design stage could be performed
using data from both pandas and spark dataframes with either
theoretical or empirical approach. Split methods support different
strategies and multigroup split. Final effect measurement stage could
be gently conducted via testing tools that allow to measure relative
and absolute effects and construct corresponding confidence intervals
for continious and binary variables. Testing tools as well as design
support significant number of statistical criteria, like t-test,
non-parametric ones, and bootstrap. For additional A/B tests support
package provides features and tools for data preproccesing and
experiment acceleration.

See "https://ambrosia.readthedocs.io" for complete documentation.

Subpackages
------------
    preprocessing - Experiment data preprocessing
    designer - Experiments design
    splitter - Groups split
    tester - Effects measurement
    tools - Core methods
    spark_tools - Spark methods
"""

from ambrosia.version import __version__



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Methods for group splitting tasks.
"""

import hashlib
import os
from base64 import b64encode
from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.manifold import TSNE

import ambrosia.tools.stratification as strat_pkg
from ambrosia import types
from ambrosia.tools.knn import NMTree

GROUPS_COLUMNS: str = "group"
AVAILLABLE_SPLIT_METHODS = ["simple", "hash", "metric", "dim_decrease"]


def check_ids_duplicates(
    dataframe: pd.DataFrame,
    id_column: Optional[types.ColumnNameType] = None,
) -> None:
    """
    Check if column with objects ids contains duplicates.
    """
    indices: np.ndarray = dataframe[id_column].values if id_column is not None else dataframe.index
    if len(indices) > len(set(indices)):
        if id_column is None:
            msg_part: str = "Index"
        else:
            msg_part: str = f"Id column {id_column}"
        raise ValueError(f"{msg_part} contains duplicates, ids must be unique for split")


def get_integer_salt(salt: Optional[str]) -> int:
    """
    Returns integer for random state numpy.

    Parameters
    ----------
    salt: Optional str

    Returns
    -------
    random_state: int
    """
    return None if salt is None else int(hashlib.shake_256(salt.encode()).hexdigest(3), 16)


def encode_id(enc_id: Any, salt: str, hash_function: Union[str, Callable] = "sha256") -> int:
    """
    Generate hash ids using prefered salt value and encoding algorithm.

    Parameters
    ----------
    enc_id : Any
        Encoding id.
    salt : str
        Salt for endoing (``enc_id`` + ``salt``).
    hash_function : str or Callable, default: ``"sha256"``
        Function that is used for id encoding.

    Returns
    -------
    enc_id_hashed : int
        Reduced hashed id.
    """
    # some constants
    LEN_HASH: int = 10  # pylint: disable=C0103
    HASH_BASE: int = 16  # pylint: disable=C0103

    enc_id = str(enc_id)
    if salt:
        enc_id += salt
    enc_id = enc_id.encode()
    hash_dict: Dict[str] = {
        "sha256": hashlib.sha256,
        "sha512": hashlib.sha512,
        "blake2": hashlib.blake2b,
    }
    if hash_function in hash_dict:
        hash_function = hash_dict[hash_function]
    elif isinstance(hash_function, str):
        raise ValueError(f"Unknown string value for hash function {', '.join(hash_dict)}")
    enc_id_hashed = hash_function(enc_id)
    if hasattr(enc_id_hashed, "hexdigest") and callable(getattr(enc_id_hashed, "hexdigest")):
        enc_id_hashed = enc_id_hashed.hexdigest()
    else:
        raise AttributeError("hexdigest() method must be implemented in hash_function")
    enc_id_hashed = enc_id_hashed[:LEN_HASH]
    enc_id_hashed = int(enc_id_hashed, HASH_BASE)
    return enc_id_hashed


def get_simple_split(
    df: pd.DataFrame, group_size: int = None, groups_number: int = 2, group_b_indices: Optional[Iterable] = None
) -> List[np.ndarray]:
    """
    Simple random uniform split approach.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe used for groups sampling.
    group_size : int
        Sampling size of groups.
    grops_number : int, default: ``2``
        Number of groups to be sampled.
    group_b_indices: np.ndarray, optional
        If group B was fixed, indices could be passed.

    Returns
    -------
    groups : List[np.ndarray]
       List of arrays with indices for each groups.
    """
    if group_b_indices is not None:
        a_indices: np.ndarray = np.random.choice(df.index.drop(group_b_indices).values, group_size, replace=False)
        return a_indices, group_b_indices
    groups: List[np.ndarray] = np.random.choice(df.index.values, size=(groups_number, group_size), replace=False)
    return groups


def get_hash_split(
    df: pd.DataFrame,
    id_column: str,
    group_size: int,
    groups_number: int = 2,
    group_b_indices: Optional[np.ndarray] = None,
    salt: Optional[str] = None,
    hash_function: Union[str, Callable] = "sha256",
) -> List[np.ndarray]:
    """
    Generate groups (basically control / test) of ids using SHA-256 hashing with salt.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe used for groups sampling.
    id_column : str
        Name of ids column in dataframe.
    group_size : int
        Sampling size of groups.
    grops_number : int, default: ``2``
        Number of groups to be sampled.
    group_b_indices: np.ndarray, optional
        If group B was fixed, indices could be passed.
    salt : str, optional
        Salt for determenistic hashing.
        If salt is not passed, a random one would be generated.
    hash_function : str or Callable, default: ``"sha256"``
        Hash function for encoding, for example sha256.
        If function is custom Callable object,
        a hexdigest() method must be implemented.

    Returns
    -------
    groups : List[np.ndarray]
       List of arrays with indices for each groups.
    """
    if group_b_indices is not None and groups_number != 2:
        raise RuntimeError(
            f"Fixed B group is available only for groups_number = 2, but groups_number = {groups_number}"
        )
    if groups_number * group_size > df.shape[0]:
        raise RuntimeError(f"Required value more, than table size {groups_number} * {group_size} > {df.shape[0]}")
    if not salt:
        salt = b64encode(os.urandom(8)).decode("ascii")
    if id_column is None:
        df_id = pd.DataFrame(df.index.values, columns=["id"])
        id_column = "id"
    else:
        df_id = df[[id_column]].copy()
    df_id.set_index(df.index.values, inplace=True)
    if group_b_indices is not None:
        df_id.drop(group_b_indices, inplace=True)
    df_id["hashed_id"] = df_id[id_column].apply(lambda x: encode_id(x, salt, hash_function)).values
    hashes: np.ndarray = np.sort(df_id["hashed_id"].values)
    groups: List[List[Any]] = []
    for group_index in range(groups_number):
        if (group_index + 1) * group_size > hashes.shape[0]:
            err_message: str = "Some problems with hash, it can be caused by collisions\n            "
            err_message += f"Required {(group_index + 1) * group_size} indexes, but hash size - {hashes.shape[0]}"
            raise RuntimeError(err_message)
        current_hashes: np.ndarray = hashes[group_index * group_size : (group_index + 1) * group_size]
        current_group: np.ndarray = df_id[df_id["hashed_id"].isin(current_hashes)].index.values
        groups.append(current_group)
        if group_b_indices is not None:
            groups.append(group_b_indices)
            break
    return groups


def get_metric_split(
    df: pd.DataFrame,
    group_size: int,
    fit_columns: List[Any],
    groups_number: int = 2,
    group_b_indices: Optional[np.ndarray] = None,
    threads: int = 1,
) -> List[np.ndarray]:
    """
    Generate groups of ids using metric approach(nearest neighbour search).

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe used for groups sampling.
    group_size : int
        Sampling size of groups.
        For strong equality choose method 'equal_size'.
    fit_columns : List
        List of columns names which values will be interpreted as
        coordinates of points in multidimensional space during metric split.
    groups_number : int, default: ``2``
        Number of groups to be sampled.
    group_b_indices : np.ndarray, optional
        If group B was fixed, indices could be passed.

    Returns
    -------
    groups : List[np.ndarray]
       List of arrays with indices for each groups.
    """
    if group_b_indices is None:
        group_b_indices = np.random.choice(df.index, size=group_size, replace=False)
    rest_index: np.ndarray = df.index.drop(group_b_indices)
    ef_search: int = group_size * (groups_number - 1)
    nm_tree = NMTree(df.loc[rest_index, fit_columns].values, payload=rest_index, ef_search=ef_search)
    result: List[List[Any]] = []
    nm_tree.query_batch(
        points=df.loc[group_b_indices, fit_columns].values,
        payload=group_b_indices,
        out=result,
        group_size=group_size,
        amount=groups_number - 1,
        threads=threads,
    )
    return np.array(result).T


def get_dim_decrease_split(
    df: pd.DataFrame,
    group_size: int,
    fit_columns: List[Any],
    groups_number: int = 2,
) -> List[np.ndarray]:
    """
    Decreased dimension split method using TSNE.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe used for groups sampling.
    group_size : int
        Sampling size of groups.
        For strong equality choose method 'equal_size'.
    fit_columns : List
        List of columns names which values will be interpreted as
        coordinates of points in multidimensional space during metric split.
    groups_number : int, default: ``2``
        Number of groups to be sampled.

    Returns
    -------
    groups : List[np.ndarray]
       List of arrays with indices for each groups.
    """
    data: np.ndarray = df[fit_columns].values
    model: TSNE = TSNE(n_components=1, learning_rate="auto", init="random", n_jobs=-2)
    labels: np.ndarray = model.fit_transform(data)
    groups: pd.DataFrame = pd.DataFrame(index=df.index)
    groups["label"] = labels
    groups.sort_values(by="label", inplace=True)
    groups = groups[: groups_number * group_size]
    result: List[np.ndarray, np.ndarray] = []
    for j in range(groups_number):
        result.append(groups.index[j::groups_number])
    return result


def make_labels_for_groups(groups_number: int) -> List[str]:
    """
    Build list with labels for groups.

    Parameters
    ----------
    groups_number : int
        Groups number for splitting.

    Returns
    -------
    List with labels len(result) = groups_number.
    """
    alphabet_size: int = ord("Z") - ord("A") + 1
    if 2 * alphabet_size < groups_number:
        raise NotImplementedError("Groups number should be <= 52")
    if groups_number <= alphabet_size:
        return [chr(ord("A") + j) for j in range(groups_number)]
    else:
        first_part: List[str] = [chr(ord("A") + j) for j in range(alphabet_size)]
        second_part: List[str] = [chr(ord("a") + j) for j in range(groups_number - alphabet_size)]
        return first_part + second_part


def add_to_required_size(
    dataframe: pd.DataFrame, required_number: int, used_ids: np.ndarray, salt: Optional[str]
) -> np.ndarray:
    """
    Finds the required number of elements from the remaining, could be stable, using salt
    Parameters
    ----------
    dataframe: pandas Dataframe
        Dataframe with all ids
    required_number: int
        Number of elements required from not used ids
    ised_ids: np.ndarray
        Ids which was used
    salt: str, Optional
        Salt for stability
    Returns
    -------
    additional_elements: np.ndarray
        Required amount new elements from dataframe
    """
    stable_random: np.random.RandomState = np.random.RandomState(get_integer_salt(salt))
    not_used_id: np.ndarray = dataframe.loc[~dataframe.index.isin(used_ids)].index.values
    additional_elements: np.ndarray = stable_random.choice(not_used_id, required_number, replace=False)
    return additional_elements


def get_split(
    dataframe: pd.DataFrame,
    split_method: str,
    groups_size: Optional[int] = None,
    groups_number: int = 2,
    group_b_indices: Iterable[int] = None,
    id_column: types.ColumnNameType = None,
    salt: Optional[str] = None,
    fit_columns: Optional[List] = None,
    strat_columns: Optional[List] = None,
    stratifier: Optional[strat_pkg.Stratification] = None,
    labels: Optional[Sequence[Any]] = None,
    threads: int = 1,
    hash_function: Union[str, Callable] = "sha256",
) -> pd.DataFrame:
    """
    Create groups split from global pool with selected set of parameters.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe used for groups sampling.
    split_method : str
        Method used for split task.
    id_column : str
        Name of ids column in dataframe.
    groups_size : int
        Sampling size of groups.
    groups_number : int, default: ``2``
        Number of groups to be sampled
    group_b_indices: np.ndarray, optional
        If group B was fixed, indices could be passed.
    fit_columns : Optional[List],
        Columns of the dataframe for metric calculation
    strat_columns : Optional[List]
        Columns for stratification https://en.wikipedia.org/wiki/Stratified_sampling
    salt : str, optional
        Salt for determenistic hashing.
        If salt is not passed, a random one would be generated.
    stratifier : Stratification, optional
        Instance of stratifier class.
    labels : Sequence[int], optional
        Labels for groups, default A, B, C ...
    threads : int, default: ``1``
        Number of threads for thread pool for metric split.
    hash_function : str or Callable, default: ``"sha256"``
        Hash function used for hash-approach, for example sha256.
        If function is custom Callable object,
        a hexdigest() method must be implemented.

    Returns
    -------
    groups : pd.DataFrame
        DataFrame with "group" column with group labels.
    """
    check_ids_duplicates(dataframe, id_column)
    if stratifier is None:
        stratifier = strat_pkg.Stratification()
        stratifier.fit(dataframe, strat_columns)
    elif not stratifier.is_trained():
        stratifier.fit(dataframe, strat_columns)

    error_size_msg: str = "Total size for all groups is bigger than total shape of table"
    cond_empty_b: bool = group_b_indices is None and groups_number * groups_size > stratifier.size()
    cond_b_id: bool = group_b_indices is not None and groups_number * len(group_b_indices) > stratifier.size()

    if cond_empty_b or cond_b_id:
        raise ValueError(error_size_msg)

    # Can't split for more than two groups with fixed test group
    if groups_number > 2 and group_b_indices is not None:
        raise NotImplementedError("Fixed group only for two groups")

    if split_method == "dim_decrease" and group_b_indices is not None:
        raise NotImplementedError("For fixed B group choose other method")

    group_ids: List[List[np.ndarray]] = [[] for _ in range(groups_number)]

    if group_b_indices is not None:
        test_inds: Dict[Tuple, Tuple[List, int]] = stratifier.get_test_inds(group_b_indices, id_column)
        groups_size = len(group_b_indices)
    elif groups_size is None:
        raise ValueError("Set groups size, if you do not set ids for B group")
    else:
        strat_sizes: Dict[Tuple, int] = stratifier.get_group_sizes(groups_size)

    for strat_value, strat_table in stratifier.groups():
        if group_b_indices is None:
            strat_group_size: int = int(strat_sizes[strat_value])
            id_b = None
        else:
            strat_group_size: int = len(test_inds[strat_value][0])
            if id_column is None:
                id_b = test_inds[strat_value][0]
            else:
                id_b = dataframe.loc[dataframe[id_column].isin(test_inds[strat_value][0])].index.values
        if split_method == "hash":
            splitted_groups = get_hash_split(
                strat_table,
                id_column=id_column,
                group_size=strat_group_size,
                groups_number=groups_number,
                group_b_indices=id_b,
                salt=salt,
                hash_function=hash_function,
            )
        elif split_method == "metric":
            splitted_groups = get_metric_split(
                strat_table,
                group_size=strat_group_size,
                fit_columns=fit_columns,
                groups_number=groups_number,
                group_b_indices=id_b,
                threads=threads,
            )
        elif split_method == "simple":
            splitted_groups = get_simple_split(
                df=strat_table, group_size=strat_group_size, groups_number=groups_number, group_b_indices=id_b
            )
        elif split_method == "dim_decrease":
            splitted_groups = get_dim_decrease_split(
                df=strat_table, group_size=strat_group_size, groups_number=groups_number, fit_columns=fit_columns
            )
        else:
            raise ValueError(f"Get incorrect split method {split_method}, choose one from {AVAILLABLE_SPLIT_METHODS}")
        for j in range(groups_number):
            group_ids[j].append(splitted_groups[j])

    # We should add new rest values for groups, which are not enough to the required size
    group_indices: List[np.ndarray] = [np.hstack(group_ids[j]) for j in range(groups_number)]
    used_ids: np.ndarray = np.hstack(group_indices)

    for j in range(groups_number):
        group: np.ndarray = group_indices[j]
        if group.shape[0] < groups_size:
            required: int = groups_size - group.shape[0]
            new_elements: np.ndarray = add_to_required_size(dataframe, required, used_ids, salt)
            group_indices[j] = np.append(group_indices[j], new_elements)
            used_ids = np.append(used_ids, new_elements)

    # Set labels
    labels = labels if labels is not None else make_labels_for_groups(groups_number)
    groups = dataframe.loc[used_ids].copy()
    groups[GROUPS_COLUMNS] = None
    for j in range(groups_number):
        groups.loc[group_indices[j], GROUPS_COLUMNS] = labels[j]
    return groups



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import enum
import typing as tp


class AmbrosiaEnum(enum.Enum):
    """
    Custom enum for Ambrosia
    """

    def _check_for_existing_members():
        """
        To implement custom enum with inheretance
        with addtional methods
        """
        pass

    @classmethod
    def check_value_in_enum(cls, value: tp.Any) -> bool:
        return value in cls._value2member_map_

    @classmethod
    def get_all_enum_values(cls) -> tp.List[str]:
        return [it.value for it in cls._member_map_.values()]

    @classmethod
    def raise_if_value_incorrect_enum(cls, value: tp.Any) -> None:
        if not cls.check_value_in_enum(value):
            msg: str = f"Choose value from " + ", ".join(cls.get_all_enum_values())
            raise ValueError(msg)


class Alternatives(AmbrosiaEnum):
    ts: str = "two-sided"
    less: str = "less"
    gr: str = "greater"


class Effects(AmbrosiaEnum):
    abs: str = "absolute"
    rel: str = "relative"
